from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Dict, List, Union

import numpy as np


@dataclass
class Code:
    """Fragment of continuous signal encoding one game state. Core class of floodlight.

    Attributes
    ----------
    code: np.ndarray
        One-dimensional array with codes describing a sequence of play.
    name: str
        Name of encoded game state (e.g. 'possession').
    definitions: dict, optional
        Dictionary of the form {token: definition} where each code category is defined
        or explained.
    framerate: int, optional
        Temporal resolution of data in frames per second/Hertz.
    token: list
        A list of all tokens used in game code, in ascending order.
    """

    code: np.ndarray
    name: str
    definitions: dict = None
    framerate: int = None

    def __str__(self):
        return f"Floodlight Code object encoding '{self.name}'"

    def __len__(self):
        return len(self.code)

    def __getitem__(self, key):
        return self.code[key]

    def __setitem__(self, key, value):
        self.code[key] = value

    def __eq__(self, other):
        return self.code == other

    def __ne__(self, other):
        return self.code != other

    def __gt__(self, other):
        return self.code > other

    def __lt__(self, other):
        return self.code < other

    def __ge__(self, other):
        return self.code >= other

    def __le__(self, other):
        return self.code <= other

    @property
    def token(self) -> list:
        token = list(np.unique(self.code))
        token.sort()

        return token

    def find_sequences(
        self, return_type: str = "dict"
    ) -> Union[Dict[Any, tuple], List[tuple]]:
        """Finds all sequences of consecutive appearances for each token and returns
        their start and end indices.

        Parameters
        ----------
        return_type: {'dict', 'list'}, default='dict'
            Specifies type of the returned sequences object.


        Returns
        -------
        sequences: Union[Dict[Any, tuple], List[tuple]]
            If ``return_type`` is 'dict', returns a dictionary of the form
            ``{token: [(sequence_start_idx, sequence_end_idx)]}``.
            If ``return_type`` is 'list', returns a list of the form
            ``[(sequence_start_idx, sequence_end_idx, token)]`` ordered by the
            respective sequence start indices.


        Examples
        --------
        >>> import numpy as np
        >>> from floodlight import Code

        >>> code = Code(code=np.array([1, 1, 2, 1, 1, 3, 1, 1]), name="intensity")
        >>> code.find_sequences()
        {1: [(0, 2), (3, 5), (6, 8)], 2: [(2, 3)], 3: [(5, 6)]}

        >>> code = Code(code=np.array(['A', 'A', 'H', 'H', 'H', 'H', 'A', 'A', 'A']),
        ...                           name="possession")
        >>> code.find_sequences(return_type="list")
        [(0, 2, 'A'), (2, 6, 'H'), (6, 9, 'A')]
        """
        if return_type not in ["dict", "list"]:
            raise ValueError(
                f"Expected return_type to be one of ['list', 'dict'], got {return_type}"
            )

        # get all token for token-wise query
        all_token = self.token
        # get change points for each token
        # NOTE: as np.diff can't be called on non-numerical token, a token-wise approach
        # is necessary, where self == token produces a boolean array
        change_points = {
            token: np.where(np.diff(self == token, prepend=np.nan, append=np.nan))[0]
            for token in all_token
        }
        # determine token-wise ranges of respective sequence lengths
        ranges = {token: range(len(change_points[token]) - 1) for token in all_token}
        # create token-wise dictionary of sequence start- and end-points
        sequences = {
            token: [
                (
                    change_points[token][i],
                    change_points[token][i + 1],
                )
                for i in ranges[token]
                if self[change_points[token][i]] == token
            ]
            for token in all_token
        }

        # combine all sequences in list and sort if desired
        if return_type == "list":
            sequences_list = []
            for token in all_token:
                sequences_list.extend(
                    [(sequence[0], sequence[1], token) for sequence in sequences[token]]
                )
            sequences = sorted(sequences_list)

        return sequences

    def slice(
        self, startframe: int = None, endframe: int = None, inplace: bool = False
    ):
        """Return copy of object with sliced code. Mimics numpy's array slicing.

        Parameters
        ----------
        startframe : int, optional
            Start of slice. Defaults to beginning of segment.
        endframe : int, optional
            End of slice (endframe is excluded). Defaults to end of segment.
        inplace: bool, optional
            If set to ``False`` (default), a new object is returned, otherwise the
            operation is performed in place on the called object.

        Returns
        -------
        code_sliced: Union[Code, None]
        """
        sliced_data = self.code[startframe:endframe].copy()
        code_sliced = None

        if inplace:
            self.code = sliced_data
        else:
            code_sliced = Code(
                code=sliced_data,
                name=deepcopy(self.name),
                definitions=deepcopy(self.definitions),
                framerate=deepcopy(self.framerate),
            )

        return code_sliced



import json
import os
from typing import Tuple
from urllib.error import HTTPError, URLError

import h5py
import numpy as np
import pandas as pd

from floodlight.io.utils import extract_zip, download_from_url
from floodlight.io.statsbomb import read_open_statsbomb_event_data_json
from floodlight import XY, Pitch, Events, Code
from settings import DATA_DIR


class EIGDDataset:
    """This dataset loads the EIGD-H data from the *A Unified Taxonomy and Multimodal
    Dataset for Events in Invasion Games* paper. [1]_

    Upon instantiation, the class checks if the data already exists in the repository's
    root ``.data``-folder, and will download the files (~120MB) to this folder if not.


    Notes
    -----
    The dataset contains a total of 25 short samples of spatiotemporal data for both
    teams and the ball from the German Men's Handball Bundesliga (HBL). For more
    information, visit the
    `official project repository <https://github.com/MM4SPA/eigd>`_.
    Data for one sample can be queried calling the :func:`~EIGDDataset.get`-method
    specifying the match and segment. The following matches and segments are
    available::

        matches = ['48dcd3', 'ad969d', 'e0e547', 'e8a35a', 'ec7a6a']
        segments = {
            '48dcd3': ['00-06-00', '00-15-00', '00-25-00', '01-05-00', '01-10-00'],
            'ad969d': ['00-00-30', '00-15-00', '00-43-00', '01-11-00', '01-35-00'],
            'e0e547': ['00-00-00', '00-08-00', '00-15-00', '00-50-00', '01-00-00'],
            'e8a35a': ['00-02-00', '00-07-00', '00-14-00', '01-05-00', '01-14-00'],
            'ec7a6a': ['00-30-00', '00-53-00', '01-19-00', '01-30-00', '01-40-00'],
            }

    Examples
    --------
    >>> from floodlight.io.datasets import EIGDDataset

    >>> dataset = EIGDDataset()
    # get one sample
    >>> teamA, teamB, ball = dataset.get(match_name="48dcd3", segment="00-06-00")
    # get the corresponding pitch
    >>> pitch = dataset.get_pitch()


    References
    ----------
        .. [1] `Biermann, H., Theiner, J., Bassek, M., Raabe, D., Memmert, D., & Ewerth,
            R. (2021, October). A Unified Taxonomy and Multimodal Dataset for Events in
            Invasion Games. In Proceedings of the 4th International Workshop on
            Multimedia Content Analysis in Sports (pp. 1-10).
            <https://dl.acm.org/doi/abs/10.1145/3475722.3482792>`_
    """

    def __init__(self, dataset_path="eigd_dataset"):
        self._EIGD_SCHEMA = "https"
        self._EIGD_BASE_URL = (
            "data.uni-hannover.de/dataset/8ccb364e-145f-4b28-8ff4-954b86e9b30d/"
            "resource/fd24e032-742d-4609-9052-cec310a2a563/download"
        )
        self._EIGD_FILENAME = "eigd-h_pos.zip"
        self._EIGD_HOST_URL = (
            f"{self._EIGD_SCHEMA}://{self._EIGD_BASE_URL}/{self._EIGD_FILENAME}"
        )
        self._EIGD_FILE_EXT = "h5"
        self._EIGD_FRAMERATE = 30

        self._data_dir = os.path.join(DATA_DIR, dataset_path)

        if not os.path.isdir(self._data_dir):
            os.makedirs(self._data_dir, exist_ok=True)
        if not bool(os.listdir(self._data_dir)):
            self._download_and_extract()

    def get(
        self, match_name: str = "48dcd3", segment: str = "00-06-00"
    ) -> Tuple[XY, XY, XY]:
        """Get one sample from the EIGD dataset.

        Parameters
        ----------
        match_name : str, optional
            Match name, check Notes section for valid arguments.
            Defaults to the first match ("48dcd3").
        segment : str, optional
            Segment identifier, check Notes section for valid arguments.
            Defaults to the first segment ("00-06-00").

        Returns
        -------
        sample: Tuple[XY, XY, XY]
            Returns three XY objects of the form (teamA, teamB, ball)
            for the requested sample.
        """
        file_name = os.path.join(
            self._data_dir, f"{match_name}_{segment}.{self._EIGD_FILE_EXT}"
        )

        if not os.path.isfile(file_name):
            raise FileNotFoundError(
                f"Could not load file, check class description for valid match "
                f"and segment values ({file_name})."
            )

        # extract from file
        with h5py.File(file_name) as h5f:
            pos_dict = {pos_set: positions[()] for pos_set, positions in h5f.items()}

        # assemble
        sample = (
            XY(xy=self._transform(pos_dict["team_a"]), framerate=self._EIGD_FRAMERATE),
            XY(xy=self._transform(pos_dict["team_b"]), framerate=self._EIGD_FRAMERATE),
            XY(xy=self._transform(pos_dict["balls"]), framerate=self._EIGD_FRAMERATE),
        )

        return sample

    @staticmethod
    def get_pitch() -> Pitch:
        """Returns a Pitch object corresponding to the EIGD-data."""
        return Pitch.from_template("eigd")

    @staticmethod
    def _transform(data: np.ndarray) -> np.ndarray:
        """Transforms spatiotemporal data from EIGD-format to floodlight format.

        Parameters
        ----------
        data: np.ndarray
            Array of shape (T, N, xydim), with T time dimension, N the number of players
            and xydim a separate dimension for x-, y-, and z-coordinates (ball only).

        Returns
        -------
        data_transformed: np.ndarray
            Array of shape (T, N*2), with T time dimension and N the number of players.
            All z-coordinates are omitted to match typical floodlight format.
        """
        # EIDG data is stored in 3-dimensional array, extract size and reshape
        T, N, _ = data.shape
        data_transformed = data[:, :, :2].reshape((T, N * 2))

        return data_transformed

    def _download_and_extract(self) -> None:
        """Downloads an archive file into temporary storage and
        extracts the content to the file system.
        """
        file = f"{DATA_DIR}/eigd.zip"
        with open(file, "wb") as binary_file:
            binary_file.write(download_from_url(self._EIGD_HOST_URL))
        extract_zip(file, self._data_dir)
        os.remove(file)


class ToyDataset:
    """This dataset loads synthetic data for a (very) short artificial football match.

    The data can be used for testing or trying out features. They come shipped with the
    package and are stored in the repository's root ``.data``-folder.

    Examples
    --------
    >>> from floodlight.io.datasets import ToyDataset

    >>> dataset = ToyDataset()
    # get one sample
    >>> (
    >>>     xy_home,
    >>>     xy_away,
    >>>     xy_ball,
    >>>     events_home,
    >>>     events_away,
    >>>     possession,
    >>>     ballstatus,
    >>> ) = dataset.get(segment="HT1")
    # get the corresponding pitch
    >>> pitch = dataset.get_pitch()

    """

    def __init__(self, dataset_path="toy_dataset"):
        self._TOY_FRAMERATE = 5
        self._TOY_DIRECTIONS = {
            "HT1": {"Home": "rl", "Away": "lr"},
            "HT2": {"Home": "lr", "Away": "rl"},
        }
        self._data_dir = os.path.join(DATA_DIR, dataset_path)

    def get(
        self, segment: str = "HT1"
    ) -> Tuple[XY, XY, XY, Events, Events, Code, Code]:
        """Get data objects for one segment from the toy dataset.

        Parameters
        ----------
        segment : str, optional
            Segment identifier for the first ("HT1") or the second ("HT2") half.
            Defaults to the first half ("HT1").

        Returns
        -------
        toy_dataset:  Tuple[XY, XY, XY, Events, Events, Code, Code]
            Returns seven XY objects of the form (xy_home, xy_away, xy_ball,
            events_home, events_away, possession, ballstatus) for the requested segment.
        """

        if segment not in ["HT1", "HT2"]:
            raise FileNotFoundError(
                f"Expected segment to be of 'HT1' or 'HT2', got {segment}"
            )

        xy_home = XY(
            xy=np.load(os.path.join(self._data_dir, f"xy_home_{segment.lower()}.npy")),
            framerate=self._TOY_FRAMERATE,
            direction=self._TOY_DIRECTIONS[segment]["Home"],
        )

        xy_away = XY(
            xy=np.load(os.path.join(self._data_dir, f"xy_away_{segment.lower()}.npy")),
            framerate=self._TOY_FRAMERATE,
            direction=self._TOY_DIRECTIONS[segment]["Away"],
        )

        xy_ball = XY(
            xy=np.load(os.path.join(self._data_dir, f"xy_ball_{segment.lower()}.npy")),
            framerate=self._TOY_FRAMERATE,
        )

        events_home = Events(
            events=pd.read_csv(
                os.path.join(self._data_dir, f"events_home_{segment.lower()}.csv")
            )
        )

        events_away = Events(
            events=pd.read_csv(
                os.path.join(self._data_dir, f"events_away_{segment.lower()}.csv")
            )
        )

        possession = Code(
            code=np.load(
                os.path.join(self._data_dir, f"possession_{segment.lower()}.npy")
            ),
            name="possession",
            definitions={1: "Home", 2: "Away"},
            framerate=self._TOY_FRAMERATE,
        )

        ballstatus = Code(
            code=np.load(
                os.path.join(self._data_dir, f"ballstatus_{segment.lower()}.npy")
            ),
            name="ballstatus",
            definitions={0: "Dead", 1: "Alive"},
            framerate=self._TOY_FRAMERATE,
        )

        data_objects = (
            xy_home,
            xy_away,
            xy_ball,
            events_home,
            events_away,
            possession,
            ballstatus,
        )

        return data_objects

    @staticmethod
    def get_pitch() -> Pitch:
        """Returns a Pitch object corresponding to the Toy Dataset."""
        return Pitch(
            xlim=(-52.5, 52.5),
            ylim=(-34, 34),
            unit="m",
            boundaries="flexible",
            length=105,
            width=68,
            sport="football",
        )


class StatsBombOpenDataset:
    """This dataset loads the StatsBomb open data provided by the `official
    data repository <https://github.com/statsbomb/open-data>`_.

    Due to the size of the full dataset (~5GB), only metadata (~2MB) are downloaded
    to the repository's root ``.data``-folder upon instantiation while the other data
    are only downloaded on demand. All downloaded files stay on disk if not manually
    removed.

    Notes
    -----
    The dataset contains results, lineups, event data, and (partly) `StatsBomb360 data
    <https://statsbomb.com/articles/soccer/
    statsbomb-360-freeze-frame-viewer-a-new-release-in-statsbomb-iq/>`_ for a variety
    of matches from a total of eight different competitions (Women's World Cup,
    FIFA World Cup, UEFA Euro, Champions League, FA Women's Super League, NWSL,
    Premier League, and La Liga).
    The Champions League data for example contains all Finals from 2003/2004 to
    2018/2019.
    The La Liga data contains every one of the 520 matches ever played by Lionel Messi
    for FC Barcelona.
    The UEFA Euro data contains 51 matches where StatsBomb360 data is available.
    As the data is constantly updated, we provide an overview over the stats here but
    refer to the official repository for up-to-date information (last
    modified 12.05.2022)::

        number_of_matches = {
            "Champions League": {
                '1999/2000' : 0, '2003/2004' : 1, '2004/2005' : 1, '2006/2007' : 1,
                '2008/2009' : 1, '2009/2010' : 1, '2010/2011' : 1, '2011/2012' : 1,
                '2012/2013' : 1, '2013/2014' : 1, '2014/2015' : 1, '2015/2016' : 1,
                '2016/2017' : 1, '2017/2018' : 1, '2018/2019' : 1,
                },
            "FA Women's Super League": {
                '2018/2019' : 108, '2019/2020' : 87, '2020/2021' : 131,
                },
            "FIFA World Cup": {
                '2018' : 64,
                },
            "La Liga": {
                '2004/2005': 7, '2005/2006' : 17, '2006/2007' : 26, '2007/2008' : 28,
                '2008/2009' : 31, '2009/2010' : 35, '2010/2011' : 33, '2011/2012' : 37,
                '2012/2013' : 32, '2013/2014' : 31, '2014/2015' : 38, '2015/2016' : 33,
                '2016/2017' : 34, '2017/2018' : 36, '2018/2019' : 34, '2019/2020' : 33,
                '2020/2021' : 35,
                },
            "NWSL": {
                '2018' : 36,
                },
            "Premier League": {
                '2003/2004' : 33,
                },
            "UEFA Euro" : {
                '2020' : 51,
                },
            "Women's World Cup": {
                '2019' : 52,
                },
        }

    Examples
    --------
    >>> from floodlight.io.datasets import StatsBombOpenDataset
    >>> dataset = StatsBombOpenDataset()
    # get one sample of event data with StatsBomb360 data
    >>> events = dataset.get("UEFA Euro", "2020", "England vs. Germany")
    # get the corresponding pitch
    >>> pitch = dataset.get_pitch()
    # get a summary of available matches in the dataset
    >>> matches = dataset.available_matches
    # extract every La Liga Clásico played in Camp Nou by Lionel Messi
    >>> clasicos = matches[matches["match_name"] == "Barcelona vs. Real Madrid"]
    # print outcomes
    >>> for _, match in clasicos.iterrows():
    >>>     print(f"Season {match['season_name']} - Barcelona {match['score']} Real'")
    # read events to list
    >>> clasico_events = []
    >>> for _, clasico in clasicos.iterrows():
    >>>     data = dataset.get("La Liga", clasico["season_name"], clasico["match_name"])
    >>>     clasico_events.append(data)
    """

    def __init__(self, dataset_path="statsbomb_dataset"):
        # setup
        self._links_competition_to_cID = {}
        self._links_season_to_sID = {}
        self._links_match_to_mID = {}
        self._STATSBOMB_SCHEMA = "https"
        self._STATSBOMB_BASE_URL = (
            "raw.githubusercontent.com/statsbomb/open-data/master/data"
        )
        self._STATSBOMB_COMPETITIONS_FILENAME = "competitions"
        self._STATSBOMB_MATCHES_FOLDERNAME = "matches"
        self._STATSBOMB_EVENTS_FOLDERNAME = "events"
        self._STATSBOMB_THREESIXTY_FOLDERNAME = "three-sixty"
        self._STATSBOMB_FILE_EXT = ".json"

        # create data directory and check if competition info needs to be downloaded
        self._data_dir = os.path.join(DATA_DIR, dataset_path)
        self.filepath_competitions = os.path.join(
            self._data_dir,
            self._STATSBOMB_COMPETITIONS_FILENAME + self._STATSBOMB_FILE_EXT,
        )
        if not os.path.isdir(self._data_dir):
            os.makedirs(self._data_dir, exist_ok=True)
        if not os.path.exists(self.filepath_competitions):
            self._download_competition_info()

        # create matches directory and check if match info needs to be downloaded
        self._matches_data_dir = os.path.join(
            self._data_dir, self._STATSBOMB_MATCHES_FOLDERNAME
        )
        if not os.path.isdir(self._matches_data_dir):
            os.makedirs(self._matches_data_dir, exist_ok=True)
        self._download_matches_info()

        # create events location
        self._events_data_dir = os.path.join(
            self._data_dir, self._STATSBOMB_EVENTS_FOLDERNAME
        )
        if not os.path.isdir(self._events_data_dir):
            os.makedirs(self._events_data_dir, exist_ok=True)

        # create StatsBomb360 location
        self._threesixty_data_dir = os.path.join(
            self._data_dir, self._STATSBOMB_THREESIXTY_FOLDERNAME
        )
        if not os.path.isdir(self._threesixty_data_dir):
            os.makedirs(self._threesixty_data_dir, exist_ok=True)

        # read links from files and update class level dictionaries
        self._read_competition_links_from_file()

    @property
    def available_matches(self) -> pd.DataFrame:
        """Creates and returns a DataFrame with information for all available matches
        from the metadata that is downloaded upon instantiation.

        Returns
        -------
        summary: pd.DataFrame
            Table where the rows contain meta information of individual games such as
            ``competition_name``, ``season_name``, and ``match_name`` (in the format
            Home vs. Away), location of the match (``stadium`` and ``country``),
            ``sex`` of the players (female or male), the ``StatsBomb360_status``  and
            the final ``score``.
        """
        summary = pd.DataFrame()

        # loop over season and competition
        for competition in self._links_competition_to_cID:
            cID = self._links_competition_to_cID[competition]
            self._read_season_match_links_for_competition_from_files(competition)
            for season in self._links_season_to_sID[competition]:
                sID = self._links_season_to_sID[competition][season]

                # loop over matches
                filepath_matches = os.path.join(
                    os.path.join(self._matches_data_dir, str(cID)),
                    str(sID) + self._STATSBOMB_FILE_EXT,
                )
                with open(filepath_matches, "r", encoding="utf8") as f:
                    matchinfo_list = json.load(f)

                for info in matchinfo_list:
                    match_info = {
                        "competition_name": competition,
                        "season_name": season,
                        "match_name": f"{info['home_team']['home_team_name']} "
                        f"vs. "
                        f"{info['away_team']['away_team_name']}",
                        "score": f"{info['home_score']}:{info['away_score']}",
                        "stadium": info["stadium"]["name"],
                        "country": info["stadium"]["country"]["name"],
                        "sex": "f"
                        if competition
                        in ["FA Women's Super League", "NWSL", "Women's World Cup"]
                        else "m",
                        "StatsBomb360_status": info["match_status_360"],
                        "cID": cID,
                        "sID": sID,
                        "mID": info["match_id"],
                    }
                    summary = summary.append(match_info, ignore_index=True)
        return summary

    def get(
        self,
        competition_name: str = "La Liga",
        season_name: str = "2020/2021",
        match_name: str = None,
    ) -> Tuple[Events, Events, Events, Events]:
        """Get events from one match of the StatsBomb open dataset.

        If `StatsBomb360data <https://statsbomb.com/articles/soccer/
        statsbomb-360-freeze-frame-viewer-a-new-release-in-statsbomb-iq/>`_  are
        available, they are stored in the  ``qualifier`` column of the Events object.
        If the files are not contained in the repository's root ``.data`` folder they
        are downloaded to the folder and will be stored until removed by hand.

        Parameters
        ----------
        competition_name : str, optional
            Competition name for which the match is played, check Notes section for
            possible competitions. Defaults to "La Liga".
        season_name : str, optional
            Season name during which the match is played. For league matches use the
            format YYYY/YYYY and for international cup matches the format YYYY.
            Check Notes for available seasons of every competition.
            Defaults to "2020/2021".
        match_name: str, optional
            Match name relating to the available matches in the chosen competition and
            season. If equal to None (default), the first available match of the
            given competition and season is chosen.

        Returns
        -------
        data_objects: Tuple[Events, Events, Events, Events]
            Returns four Events objects of the form (events_home_ht1, events_home_ht2,
            events_away_ht1, events_away_ht2) for the requested sample.
        """
        # get identifiers from links
        cID = self._links_competition_to_cID[competition_name]
        if competition_name not in self._links_season_to_sID:
            self._read_season_match_links_for_competition_from_files(competition_name)
        sID = self._links_season_to_sID[competition_name][season_name]
        matches_dict = self._links_match_to_mID[competition_name][season_name]
        if match_name is None:
            mID = list(matches_dict.values())[0]
        else:
            mID = matches_dict[match_name]

        # create paths
        filepath_matches = os.path.join(
            os.path.join(self._matches_data_dir, str(cID)),
            str(sID) + self._STATSBOMB_FILE_EXT,
        )
        filepath_events = os.path.join(
            self._events_data_dir,
            str(mID) + self._STATSBOMB_FILE_EXT,
        )
        filepath_threesixty = os.path.join(
            self._threesixty_data_dir,
            str(mID) + self._STATSBOMB_FILE_EXT,
        )

        # check if events need to be downloaded
        if not os.path.exists(filepath_events):
            events_host_url = (
                f"{self._STATSBOMB_SCHEMA}://"
                f"{self._STATSBOMB_BASE_URL}/"
                f"{self._STATSBOMB_EVENTS_FOLDERNAME}/"
                f"{str(mID)}"
                f"{self._STATSBOMB_FILE_EXT}"
            )
            with open(filepath_events, "wb") as binary_file:
                binary_file.write(download_from_url(events_host_url))

        # check if StatsBomb360 data is available and needs to be downloaded
        if not os.path.exists(filepath_threesixty):
            threesixty_host_url = (
                f"{self._STATSBOMB_SCHEMA}://"
                f"{self._STATSBOMB_BASE_URL}/"
                f"{self._STATSBOMB_THREESIXTY_FOLDERNAME}/"
                f"{str(mID)}"
                f"{self._STATSBOMB_FILE_EXT}"
            )
            try:
                data = download_from_url(threesixty_host_url)
                with open(filepath_threesixty, "wb") as binary_file:
                    binary_file.write(data)
            except HTTPError:
                filepath_threesixty = None

        # read events from file
        (home_ht1, home_ht2, away_ht1, away_ht2,) = read_open_statsbomb_event_data_json(
            filepath_events, filepath_matches, filepath_threesixty
        )
        event_objects = (home_ht1, home_ht2, away_ht1, away_ht2)

        return event_objects

    @staticmethod
    def get_pitch() -> Pitch:
        """Returns a Pitch object corresponding to the StatsBomb Dataset."""
        return Pitch.from_template("statsbomb", sport="football")

    def _read_competition_links_from_file(self):
        """Writes the data links between the available competitions and the respective
        cIDs to the class level dictionary.
        """
        # updates on competition level
        competition_info = pd.read_json(self.filepath_competitions)
        cIDs = competition_info["competition_id"].unique()
        competitions = competition_info["competition_name"].unique()
        self._links_competition_to_cID.update(
            {competition: cIDs[i] for i, competition in enumerate(competitions)}
        )

    def _read_season_match_links_for_competition_from_files(self, competition_name):
        """Writes data links between the seasons and matches to the respective sIDs
        and mIDs for a given competition to the class level dictionaries.
        """
        # read competition file
        cID = self._links_competition_to_cID[competition_name]
        competition_info = pd.read_json(self.filepath_competitions)

        # update season and match dictionaries with competition information
        self._links_season_to_sID.update({competition_name: {}})
        self._links_match_to_mID.update({competition_name: {}})

        # loop over all available seasons of the given competition
        for _, single_season in competition_info.iterrows():
            if cID != single_season["competition_id"]:
                continue

            # update season and match dictionaries with season information
            sID = single_season["season_id"]
            season_name = single_season["season_name"]
            self._links_season_to_sID[competition_name].update({season_name: sID})
            self._links_match_to_mID[competition_name].update({season_name: {}})

            # read information of all matches within the season
            filepath_matches = os.path.join(
                os.path.join(self._matches_data_dir, str(cID)),
                str(sID) + self._STATSBOMB_FILE_EXT,
            )
            with open(filepath_matches, "rb") as matches_file:
                season_matches_info = json.load(matches_file)

            # update match dictionary with match information
            for info in season_matches_info:
                match_name = (
                    f"{info['home_team']['home_team_name']} vs. "
                    f"{info['away_team']['away_team_name']}"
                )
                mID = info["match_id"]
                self._links_match_to_mID[competition_name][season_name].update(
                    {match_name: mID}
                )

    def _download_competition_info(self) -> None:
        """Downloads json file with competition information into the file system."""
        competitions_host_url = (
            f"{self._STATSBOMB_SCHEMA}://"
            f"{self._STATSBOMB_BASE_URL}/"
            f"{self._STATSBOMB_COMPETITIONS_FILENAME}"
            f"{self._STATSBOMB_FILE_EXT}"
        )

        # download file with information of all seasons
        try:
            with open(self.filepath_competitions, "wb") as binary_file:
                binary_file.write(download_from_url(competitions_host_url))
        except URLError:  # remove empty json file if download fails
            os.remove(self.filepath_competitions)
            raise URLError(
                f"Could not download competitions.json from URL "
                f"{competitions_host_url}. Check your internet connection!"
            )

    def _download_matches_info(self) -> None:
        """Downloads the json files containing information about available matches from
        all available seasons into the file system.
        """
        competition_info = pd.read_json(self.filepath_competitions)

        for _, single_season in competition_info.iterrows():
            cID = single_season["competition_id"]
            sID = single_season["season_id"]
            matches_filepath = os.path.join(
                os.path.join(self._matches_data_dir, str(cID)),
                str(sID) + self._STATSBOMB_FILE_EXT,
            )
            if not os.path.exists(matches_filepath):
                season_host_url = (
                    f"{self._STATSBOMB_SCHEMA}://"
                    f"{self._STATSBOMB_BASE_URL}/"
                    f"{self._STATSBOMB_MATCHES_FOLDERNAME}/"
                    f"{str(cID)}/"
                    f"{str(sID)}"
                    f"{self._STATSBOMB_FILE_EXT}"
                )
                competition_data_dir = os.path.join(self._matches_data_dir, str(cID))
                if not os.path.isdir(competition_data_dir):
                    os.makedirs(competition_data_dir, exist_ok=True)
                season_file = os.path.join(
                    competition_data_dir, str(sID) + self._STATSBOMB_FILE_EXT
                )
                with open(season_file, "wb") as binary_file:
                    binary_file.write(download_from_url(season_host_url))



from .core.code import Code
from .core.events import Events
from .core.pitch import Pitch
from .core.xy import XY

__all__ = ["__version__", "__doc__", "Code", "Events", "Pitch", "XY"]

__version__ = "0.3.0"

__doc__ = """
A high-level, data-driven sports analytics framework
====================================================

**floodlight** is a Python package for streamlined analysis of sports data. It is
designed with a clear focus on scientific computing and built upon popular libraries
such as *numpy* or *pandas*.

Load, integrate, and process tracking and event data, codes and other match-related
information from major data providers. This package provides a set of  standardized
data objects to structure and handle sports data, together with a suite of common
processing operations such as transforms or data manipulation methods.

All implementations run completely provider- and sports-independent, while maintaining
a maximum of flexibility to incorporate as many data flavours as possible. A high-level
interface allows easy access to all standard routines, so that you can stop worrying
about data wrangling and start focussing on the analysis instead!
"""



import numpy as np
import datetime


# fmt: off
essential_events_columns = {
    "eID": {
        "definition": "Event ID - unique number or string that identifies the event "
                      "type. The resulting system can either be provider specific, or "
                      "customary. However, a link between eID and the event "
                      "definitions/descriptions should be available",
        "dtypes": [str, int],
        "value_range": None,
    },
    "gameclock": {
        "definition": "Elapsed time relative to segment start in seconds",
        "dtypes": [float],
        "value_range": [0, np.inf]
    }
}


protected_columns = {
    "pID": {
        "definition": "Player ID - unique number or string for player identification ",
        "dtypes": [str, int],
        "value_range": None
    },
    "jID": {
        "definition": "Jersey ID - a players jersey number in a single observation",
        "dtypes": [int],
        "value_range": [0, np.inf]
    },
    "xID": {
        "definition": "Index ID - a players index in the list of all players of a team"
                      "for a given match (starts counting at 1). This is primarily used"
                      " for locating players data in XY objects, but can also be "
                      "helpful iterating or displaying all players of a team",
        "dtypes": [int],
        "value_range": [1, np.inf]
    },
    "tID": {
        "definition": "Team ID - unique number or string for team identification ",
        "dtypes": [str, int],
        "value_range": None
    },
    "mID": {
        "definition": "Match ID - unique number or string for match identification ",
        "dtypes": [str, int],
        "value_range": None
    },
    "cID": {
        "definition": "Competition ID - unique number or string for competition (e.g. "
                      "league or cup) identification",
        "dtypes": [str, int],
        "value_range": None
    },
    "frameclock": {
        "definition": "Elapsed time relative to segment start in frames given a certain"
                      "framerate.",
        "dtypes": [int],
        "value_range": [0, np.inf]
    },
    "timestamp": {
        "definition": "Datetime timestamp. Should be aware and carry a pytz timezone",
        "dtypes": [datetime.datetime],
        "value_range": None
    },
    "minute": {
        "definition": "Minute of the segment the event took place",
        "dtypes": [int],
        "value_range": [0, np.inf]
    },
    "second": {
        "definition": "Second of the minute of the segment the event took place",
        "dtypes": [int],
        "value_range": [0, np.inf]
    },
    "outcome": {
        "definition": "Result of an event as included by many data providers. "
                      "Positive/Successful is 1, Negative/Unsuccessful is 0",
        "dtypes": [int],
        "value_range": [0, 1],
    },
    "at_x": {
        "definition": "The x position (longitudinal) where the event took place or "
                      "originated from",
        "dtypes": [float],
        "value_range": None
    },
    "at_y": {
        "definition": "The y position (lateral) where the event took place or "
                      "originated from",
        "dtypes": [float],
        "value_range": None
    },
    "to_x": {
        "definition": "The x position (longitudinal) where the event ended",
        "dtypes": [float],
        "value_range": None
    },
    "to_y": {
        "definition": "The y position (lateral) where the event ended",
        "dtypes": [float],
        "value_range": None
    },
}

# fmt:on



from floodlight.io.kinexon import read_kinexon_file

from floodlight import Pitch

from floodlight.models.geometry import CentroidModel#, StretchIndexModel, PlayerDistanceModel


# PATHS
sample_kinexon_path = "C:\\Users\\Dominik\\sciebo\\floodlight\\Kinexon\\Game 1.csv"
sample_f24_path = "C:\\Users\\Dominik\\sciebo\\floodlight\\Opta\\f24-4-2017-958085-eventdetails.xml"
sample_tracab_data_path = "C:\\Users\\Dominik\\sciebo\\floodlight\\Tracab (ChyronHego)\\1044528_shortened.dat"
sample_tracab_meta_path = "C:\\Users\\Dominik\\sciebo\\floodlight\\Tracab (ChyronHego)\\1044528_metadata.xml"


# KINEXON DATA
xy0, xy1, xy3 = read_kinexon_file(sample_kinexon_path)

pitch = Pitch((0, 40), (0, 20), 'percent', 'flexible', 40, 20, 'handball')
cm = CentroidModel()
# cm.fit(xy0)
cent = cm.centroid()
cent = cm.centroid_distance(xy0)

# sim = StretchIndexModel()
# sim.fit(xy0, cent)
# stretch = sim.get_stretch_index()
#
# pdm = PlayerDistanceModel(pitch)
# pdm.fit(xy0)
# dist = pdm.get_player_distances()
# pdm.fit(xy0, xy1)
# dist2 = pdm.get_player_distances()




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['floodlight',
 'floodlight.core',
 'floodlight.io',
 'floodlight.models',
 'floodlight.transforms',
 'floodlight.utils',
 'floodlight.vis']

package_data = \
{'': ['*']}

install_requires = \
['h5py>=3.6.0,<4.0.0',
 'iso8601>=1.0.2,<2.0.0',
 'lxml>=4.6.4,<5.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.21.2,<2.0.0',
 'pandas>=1.3.4,<2.0.0',
 'pytz>=2021.3,<2022.0',
 'scipy>=1.8.0,<2.0.0']

setup_kwargs = {
    'name': 'floodlight',
    'version': '0.3.0',
    'description': 'A high-level framework for sports data analysis',
    'long_description': "[version-image]: https://img.shields.io/pypi/v/floodlight?color=006666\n[version-url]: https://pypi.org/project/floodlight/\n[python-image]: https://img.shields.io/pypi/pyversions/floodlight?color=006666\n[python-url]: https://pypi.org/project/floodlight/\n[docs-image]: https://readthedocs.org/projects/floodlight/badge/?version=latest\n[docs-url]: https://floodlight.readthedocs.io/en/latest/?badge=latest\n[build-image]: https://github.com/floodlight-sports/floodlight/actions/workflows/build.yaml/badge.svg\n[build-url]: https://github.com/floodlight-sports/floodlight/actions/workflows/build.yaml\n[lint-image]: https://github.com/floodlight-sports/floodlight/actions/workflows/linting.yaml/badge.svg\n[lint-url]: https://github.com/floodlight-sports/floodlight/actions/workflows/linting.yaml\n[black-image]: https://img.shields.io/badge/code%20style-black-000000.svg\n[black-url]: https://github.com/psf/black\n[contrib-image]: https://img.shields.io/badge/contributions-welcome-006666\n[contrib-url]: https://github.com/floodlight-sports/floodlight/blob/main/CONTRIBUTING.md\n\n# floodlight\n[![Latest Version][version-image]][version-url]\n[![Python Version][python-image]][python-url]\n[![Documentation Status][docs-image]][docs-url]\n[![Build Status][build-image]][build-url]\n[![Linting Status][lint-image]][lint-url]\n[![Contributions][contrib-image]][contrib-url]\n[![Code style: black][black-image]][black-url]\n\n\n## A high-level, data-driven sports analytics framework\n\n**floodlight** is a Python package for streamlined analysis of sports data. It is\ndesigned with a clear focus on scientific computing and built upon popular libraries\nsuch as *numpy* or *pandas*.\n\nLoad, integrate, and process tracking and event data, codes and other match-related\ninformation from major data providers. This package provides a set of  standardized\ndata objects to structure and handle sports data, together with a suite of common\nprocessing operations such as transforms or data manipulation methods.\n\nAll implementations run completely provider- and sports-independent, while maintaining\na maximum of flexibility to incorporate as many data flavours as possible. A high-level\ninterface allows easy access to all standard routines, so that you can stop worrying\nabout data wrangling and start focussing on the analysis instead!\n\n----------------------------------------------------------------------------------------\n\n* [Features](#features)\n* [Installation](#installation)\n* [Documentation](#documentation)\n* [How to contribute](#contributing)\n\n----------------------------------------------------------------------------------------\n\n\n### Features\n\nThis project is still under development, and we hope to expand the set\nof features in the future. At this point, we provide core data structures,\nparsing functionality for major data providers, access to public data sets, data\nfiltering, basic plotting routines and computational models.\n\n#### Data-level Objects\n\n- Tracking data\n- Event data\n- Pitch information\n- Codes such as ball possession information\n- Properties such as distances or advanced computations\n\n#### Parser\n\n- ChyronHego (Tracking data, Codes)\n- DFL (Tracking data, Event data, Codes)\n- Kinexon (Tracking data)\n- Opta (Event data - F24 feeds)\n- StatsPerform (Tracking data, Event data - also directly from URLs)\n- StatsBomb (Event data)\n\n#### Datasets\n\n- EIGD-H (Handball tracking data)\n- StatsBomb OpenData (Football event data)\n\n#### Manipulation and Plotting\n\n- Spatial transformations for all data structures\n- Lowpass-filter tracking data\n- Slicing and selection methods\n- Plot pitches and tracking data\n\n#### Models\n\n- Centroids\n- Distances, Velocities, Accelerations\n- Metabolic Power and Equivalent Distances\n\n### Installation\n\nThe package can be installed easily via pip:\n\n```\npip install floodlight\n```\n\n\n### Documentation\n\nYou can find all documentation [here][docs-url].\n\n\n\n### Contributing\n\nCheck out [Contributing.md][contrib-url] for a quick rundown of what you need to\nknow to get started. We also provide an extended, beginner-friendly guide on how to\nstart contributing in our documentation.\n\n\n\n### Why\n\nWhy do we need another package that introduces its own data structures and ways of dealing with certain problems? And,\nto be honest, what's the purpose of trying to integrate all these different files and fit them into a single framework?\nEspecially since there already exist packages that aim to solve certain parts of that pipeline?\n\nThe answer is, while we love those packages out there, that we did not find a solution that did fit our needs.\nAvailable packages are either tightly connected to a certain data format/provider, adapt to the subtleties of a\nparticular sport, or only solve *one* particular problem. This still left us with the essential problem of adapting to\nall those different interfaces.\n\nWe felt that as long as there is no underlying, high-level framework, each and every use case again and again needs its\nown implementation. At last, we found ourselves refactoring the same code - and there are certain processing or\nplotting routines that are required in *almost every* project - over and over again, just to fit the particular data\nstructures we were dealing with at that time.\n\n\n### About\n\nThis project has been kindly supported by the [Institute of Exercise Training and Sport\nInformatics](https://www.dshs-koeln.de/en/institut-fuer-trainingswissenschaft-und-sportinformatik/) at the German Sport\nUniversity Cologne under supervision of Prof. Daniel Memmert.\n\n\n\n### Related Projects\n\n- [matplotsoccer](https://github.com/TomDecroos/matplotsoccer)\n- [kloppy](https://github.com/PySport/kloppy)\n- [codeball](https://github.com/metrica-sports/codeball)\n",
    'author': 'draabe',
    'author_email': 'draabx@posteo.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/floodlight-sports/floodlight',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import Dict, Tuple, Union
import warnings

from lxml import etree
import iso8601
import numpy as np
import pandas as pd

from floodlight.core.code import Code
from floodlight.core.events import Events
from floodlight.core.pitch import Pitch
from floodlight.core.xy import XY


def _create_periods_from_dat(
    filepath_positions: Union[str, Path]
) -> Tuple[Dict[str, Tuple[int, int]], int]:
    """Parses over position file and returns dictionary with periods as well as an
    estimate of the framerate based on the timedelta between multiple frames.

    Parameters
    ----------
    filepath_positions: str or pathlib.Path
        Path to XML File where the Position data in DFL format is saved.

    Returns
    -------
    periods: Dict[str, Tuple[int, int]
        Dictionary with times for the segment:
        ``periods[segment] = (starttime, endtime)``
    est_framerate: int
        Estimated temporal resolution of data in frames per second/Hertz.
    """
    periods = {}
    framerate_est = None

    # retrieve information from ball frame sets
    for _, frame_set in etree.iterparse(filepath_positions, tag="FrameSet"):
        if frame_set.get("TeamId").lower() == "ball":
            frames = [frame for frame in frame_set.iterfind("Frame")]
            periods[frame_set.get("GameSection")] = (
                int(frames[0].get("N")),
                int(frames[-1].get("N")),
            )
            delta = iso8601.parse_date(frames[1].get("T")) - iso8601.parse_date(
                frames[0].get("T")
            )
            if framerate_est is None:
                framerate_est = int(round(1 / delta.total_seconds()))
            elif framerate_est != int(round(1 / delta.total_seconds())):
                warnings.warn(
                    f"Framerate estimation yielded diverging results."
                    f"The originally estimated framerate of {framerate_est} Hz did not "
                    f"match the current estimation of "
                    f"{int(round(1 / delta.total_seconds()))} Hz. This might be "
                    f"caused by missing frame(s) in the position data."
                    f"Continuing by choosing the latest estimation of "
                    f"{int(round(1 / delta.total_seconds()))} Hz"
                )
                framerate_est = int(round(1 / delta.total_seconds()))

        frame_set.clear()

    return periods, framerate_est


def create_links_from_mat_info(
    filepath_mat_info: Union[str, Path]
) -> Tuple[Dict[str, Dict[int, int]], Dict[str, Dict[str, int]]]:
    """Parses the DFL Match Information XML file for unique jIDs (jerseynumbers) and
    creates two dictionaries, one linking pIDs to jIDs and one linking jIDs to xIDs in
    ascending order.

    Parameters
    ----------
    filepath_mat_info: str or pathlib.Path
        Full path to XML File where the Match Information data in DFL format is saved

    Returns
    -------
    links_jID_to_xID: Dict[str, Dict[int, int]]
        A link dictionary of the form `links[team][jID] = xID`.
    links_pID_to_jID: Dict[str, Dict[str, int]]
        A link dictionary of the form `links[team][pID] = jID`.
    """
    # set up XML tree
    tree = etree.parse(str(filepath_mat_info))
    root = tree.getroot()

    # parse XML file and extract links from pID to xID for both teams
    links_pID_to_jID = {}
    teams = root.find("MatchInformation").find("Teams")
    home = root.find("MatchInformation").find("General").get("HomeTeamId")
    if "AwayTeamId" in root.find("MatchInformation").find("General").attrib:
        away = root.find("MatchInformation").find("General").get("AwayTeamId")
    elif "GuestTeamId" in root.find("MatchInformation").find("General").attrib:
        away = root.find("MatchInformation").find("General").get("GuestTeamId")
    else:
        away = None

    for team in teams:
        if team.get("TeamId") == home:
            links_pID_to_jID["Home"] = {
                player.get("PersonId"): int(player.get("ShirtNumber"))
                for player in team.find("Players")
            }
        elif team.get("TeamId") == away:
            links_pID_to_jID["Away"] = {
                player.get("PersonId"): int(player.get("ShirtNumber"))
                for player in team.find("Players")
            }
        else:
            continue

    links_jID_to_xID = {
        "Home": {
            int(links_pID_to_jID["Home"][pID]): xID
            for xID, pID in enumerate(links_pID_to_jID["Home"])
        },
        "Away": {
            int(links_pID_to_jID["Away"][pID]): xID
            for xID, pID in enumerate(links_pID_to_jID["Away"])
        },
    }

    return links_jID_to_xID, links_pID_to_jID


def _get_event_description(
    elem: etree.Element,
) -> Tuple[str, Dict[str, Union[str, int]]]:
    """Returns the full description of a single XML Event in the DFL format.

    Parameters
    ----------
    elem: lxml.etree.Element
        lxml.etree.Element with the Event information.

    Returns
    -------
    eID: str
        High-level description for the current event.
    attrib: Dict
        Additional attributes for the current event in the form
        ``attrib[category] = label``.
    """
    # read description
    eID = elem.tag

    # read additional attributes
    attrib = {}
    for category in elem.attrib:
        attrib[category] = elem.attrib[category]

    # find nested elements
    nested_eID = None
    nested_attrib = None
    if elem.find("Play") is not None:
        nested_eID, nested_attrib = _get_event_description(elem.find("Play"))
    elif elem.find("Pass") is not None:
        nested_eID, nested_attrib = _get_event_description(elem.find("Pass"))
    elif elem.find("Cross") is not None:
        nested_eID, nested_attrib = _get_event_description(elem.find("Cross"))
    elif elem.find("ShotAtGoal") is not None:
        nested_eID, nested_attrib = _get_event_description(elem.find("ShotAtGoal"))
        nested_attrib.update(attrib)
    elif elem.find("SuccessfulShot") is not None:
        nested_eID, nested_attrib = _get_event_description(elem.find("SuccessfulShot"))
        nested_attrib.update(attrib)
    elif elem.find("SavedShot") is not None:
        nested_eID, nested_attrib = _get_event_description(elem.find("SavedShot"))
        nested_attrib.update(attrib)
    elif elem.find("BlockedShot") is not None:
        nested_eID, nested_attrib = _get_event_description(elem.find("BlockedShot"))
        nested_attrib.update(attrib)
    elif elem.find("ShotWide") is not None:
        nested_eID, nested_attrib = _get_event_description(elem.find("ShotWide"))
        nested_attrib.update(attrib)
    elif elem.find("ShotWoodWork") is not None:
        nested_eID, nested_attrib = _get_event_description(elem.find("ShotWoodWork"))
        nested_attrib.update(attrib)
    elif elem.find("OtherShot") is not None:
        nested_eID, nested_attrib = _get_event_description(elem.find("OtherShot"))

    # update nested elements
    if nested_eID is not None and nested_attrib is not None:
        eID += "_" + nested_eID
        attrib.update(nested_attrib)

    return eID, attrib


def _get_event_outcome(eID, attrib) -> int:
    """Returns the outcome of a single Event in the DFL format.

    Parameters
    ----------
    eID: str
        High-level description for the current event.
    attrib: Dict
        Additional attributes for the current event in the form
        ``attrib[category] = label``.

    Returns
    -------
    outcome: int
        Outcome coded as 1 (success) or 0 (failure) of the current event or np.nan in
        case no outcome is defined.
    """
    outcome = np.nan

    # well-defined outcome
    if "TacklingGame" in eID:
        if attrib["WinnerRole"] == "withoutBallControl":
            outcome = 1
        elif attrib["WinnerRole"] == "withBallControl":
            outcome = 0
    elif "BallClaiming" in eID:
        if "Type" in attrib:
            if attrib["Type"] in ["BallClaimed"]:
                outcome = 1
            elif attrib["Type"] in ["BallHeld"]:
                outcome = 0
    elif "Play" in eID:
        if "Successful" in attrib:
            if attrib["Successful"] == "true":
                outcome = 1
            elif attrib["Successful"] == "false":
                outcome = 0
    elif "ShotAtGoal" in eID:
        if "SuccessfulShot" in eID:
            outcome = 1
        else:
            outcome = 0
    # per definition no outcome
    elif eID in [
        "OwnGoal",
        "PreventedOwnGoal",
        "DefensiveClearance",
        "Foul",
        "Offside",
        "Caution",
        "SendingOff",
        "Substitution",
        "KickoffWhistle",
        "FinalWhistle",
        "FairPlay",
        "RefereeBall",
        "OtherBallAction",
        "OtherPlayerAction",
    ]:
        pass
    # missing sub child in event
    elif eID in [
        "FreeKick",
        "ThrowIn",
        "CornerKick",
        "Penalty",
        "GoalKick",
        "Kickoff",
    ]:
        pass

    return outcome


def _get_event_team_and_player(eID, attrib) -> Tuple[str, str]:
    """Returns the player and team of a single Event in the DFL format.

    Parameters
    ----------
    eID: str
        High-level description for the current event.
    attrib: Dict
        Additional attributes for the current event in the form
        ``attrib[category] = label``.

    Returns
    -------
    team: str
        DFL Team ID of the Event
    player: str
        DFL Player ID of the Event
    """
    # team
    team = None
    if "Team" in attrib:
        team = attrib["Team"]
    elif eID == "TacklingGame" and "WinnerTeam" in attrib and "LoserTeam" in attrib:
        if attrib["WinnerRole"] == "withBallControl":
            team = attrib["WinnerTeam"]
        elif attrib["WinnerRole"] == "withoutBallControl":
            team = attrib["LoserTeam"]
    elif "TeamFouler" in attrib:
        team = attrib["TeamFouler"]

    # player
    player = None
    if "Player" in attrib:
        player = attrib["Player"]
    elif eID == "TacklingGame" and "Winner" in attrib and "Loser" in attrib:
        if attrib["WinnerRole"] == "withBallControl":
            player = attrib["Winner"]
        elif attrib["WinnerRole"] == "withoutBallControl":
            player = attrib["Loser"]
    elif "Fouler" in attrib:
        player = attrib["Fouler"]

    return team, player


def read_pitch_from_mat_info_xml(filepath_mat_info: Union[str, Path]) -> Pitch:
    """Reads match_information XML file and returns the playing Pitch.

    Parameters
    ----------
    filepath_mat_info: str or pathlib.Path
        Full path to XML File where the Match Information data in DFL format is saved.

    Returns
    -------
    pitch: Pitch
        Pitch object with actual pitch length and width.
    """
    # set up XML tree
    tree = etree.parse(str(filepath_mat_info))
    root = tree.getroot()

    # parse pitch
    length = root.find("MatchInformation").find("Environment").get("PitchX")
    length = float(length) if length else None
    width = root.find("MatchInformation").find("Environment").get("PitchY")
    width = float(width) if width else None
    pitch = Pitch.from_template(
        "dfl",
        length=length,
        width=width,
        sport="football",
    )

    return pitch


def read_event_data_xml(
    filepath_events: Union[str, Path]
) -> Tuple[Events, Events, Events, Events]:
    """Parses a DFL Match Event XML file and extracts the event data.

    This function provides a high-level access to the particular DFL Match Event feed
    and returns Event objects for both teams. The number of segments is inferred from
    the data, yet data for each segment is stored in a separate object.

    Parameters
    ----------
    filepath_events: str or pathlib.Path
        Full path to XML File where the Event data in DFL format is saved.

    Returns
    -------
    data_objects: Tuple[Events, Events, Events, Events]
        Events- and Pitch-objects for both teams and both halves.

    Notes
    -----
    The DFL format of handling event data information involves an elaborate use of
    certain event attributes, which attach additional information to certain events.
    There also exist detailed definitions for these attributes. Parsing this information
    involves quite a bit of logic and is planned to be included in further releases. As
    of now, qualifier information is parsed as a string in the `qualifier` column of the
    returned DataFrame and might be transformed to a dict of the form:
    `{attribute: value}`.
    """
    # set up XML tree
    tree = etree.parse(str(filepath_events))
    root = tree.getroot()

    # find start and end of half
    kickoff_whistles = {}
    final_whistles = {}
    for whistle in root.findall("Event/KickoffWhistle"):
        if whistle.get("GameSection") is not None:
            kickoff_whistles[whistle.get("GameSection")] = iso8601.parse_date(
                whistle.getparent().get("EventTime")
            )
    for whistle in root.findall("Event/FinalWhistle"):
        if whistle.get("GameSection") is not None:
            final_whistles[whistle.get("GameSection")] = iso8601.parse_date(
                whistle.getparent().get("EventTime")
            )

    # initialize periods
    segments = list(kickoff_whistles.keys())
    periods = {}
    for segment in segments:
        periods[segment] = (kickoff_whistles[segment], final_whistles[segment])

    # set up bins
    team_events = {segment: {} for segment in segments}

    # loop over events
    for elem in root.findall("Event"):
        # initialize
        event = {}

        # check for structure that is an element Event with a single child
        if len(elem) != 1:
            warnings.warn(
                "An XML Event has multiple children. This likely causes imprecise "
                "Event descriptions and outcomes."
            )

        # absolute time information (timestamp)
        event["timestamp"] = iso8601.parse_date(elem.get("EventTime"))
        event["gameclock"] = np.nan

        # segment in which event took place
        segment = None
        for seg in segments:
            if periods[seg][0] <= event["timestamp"] <= periods[seg][1]:
                segment = seg
        # assign to closest start point if not within any segments
        if segment is None:
            seg_ind = np.argmin(
                [np.abs(event["timestamp"] - periods[seg][0]) for seg in segments]
            )
            segment = segments[int(seg_ind)]

        # relative time information (gameclock)
        event["gameclock"] = (event["timestamp"] - periods[segment][0]).total_seconds()
        event["minute"] = np.floor(event["gameclock"] / 60)
        event["second"] = np.floor(event["gameclock"] - event["minute"] * 60)

        # description, outcome, team, and player
        child = next(iter(elem))
        eID, attrib = _get_event_description(child)
        outcome = _get_event_outcome(eID, attrib)
        tID, pID = _get_event_team_and_player(eID, attrib)
        event["eID"] = eID
        event["qualifier"] = attrib
        event["outcome"] = outcome
        event["tID"] = tID
        event["pID"] = pID

        # insert to bin
        if tID not in team_events[segment]:
            team_events[segment][tID] = []
        if event["eID"] == "Substitution":  # split for the special case substitution
            # in-sub
            event["eID"] = "InSubstitution"
            event["pID"] = event["qualifier"]["PlayerIn"]
            team_events[segment][tID].append(event)
            # out-sub
            event["eID"] = "OutSubstitution"
            event["pID"] = event["qualifier"]["PlayerOut"]
            team_events[segment][tID].append(event)
        else:
            team_events[segment][tID].append(event)

    # postprocessing
    team_dfs = {segment: {} for segment in segments}
    for segment in segments:

        # teams
        teams = [tID for tID in team_events[segment] if tID is not None]

        # loop over teams
        for tID in teams:

            # assign events with tID None to both teams
            team_events[segment][tID] += team_events[segment][None]

            # transform to data DataFrame
            team_dfs[segment][tID] = pd.DataFrame(team_events[segment][tID])

            # columns to standard order
            team_dfs[segment][tID] = team_dfs[segment][tID][
                [
                    "eID",
                    "gameclock",
                    "tID",
                    "pID",
                    "outcome",
                    "timestamp",
                    "minute",
                    "second",
                    "qualifier",
                ]
            ]
            team_dfs[segment][tID] = team_dfs[segment][tID].sort_values("gameclock")
            team_dfs[segment][tID] = team_dfs[segment][tID].reset_index(drop=True)

    # check for teams
    team1 = list(team_dfs[segments[0]].keys())[0]
    team2 = list(team_dfs[segments[0]].keys())[1]
    if not np.all([team1 in team_dfs[segment].keys() for segment in segments]):
        KeyError(
            f"Found tID {team1} of the first segment missing in at least one "
            f"other segment!"
        )
    if not np.all([team2 in team_dfs[segment].keys() for segment in segments]):
        KeyError(
            f"Found tID {team2} of the first segment missing in at least one "
            f"other segment!"
        )

    # assembly
    events_team1_ht1 = Events(
        events=team_dfs[segments[0]][team1],
    )
    events_team1_ht2 = Events(
        events=team_dfs[segments[1]][team1],
    )
    events_team2_ht1 = Events(
        events=team_dfs[segments[0]][team2],
    )
    events_team2_ht2 = Events(
        events=team_dfs[segments[1]][team2],
    )
    data_objects = (
        events_team1_ht1,
        events_team1_ht2,
        events_team2_ht1,
        events_team2_ht2,
    )

    return data_objects


def read_position_data_xml(
    filepath_positions: Union[str, Path],
    filepath_mat_info: Union[str, Path],
    links_jID_to_xID: Dict[str, Dict[int, int]] = None,
    links_pID_to_jID: Dict[str, Dict[int, int]] = None,
) -> Tuple[XY, XY, XY, XY, XY, XY, Code, Code, Code, Code, Pitch]:
    """Parse DFL files and extract position data, possession and ballstatus codes as
    well as pitch information.

    The official tracking system of the DFL (German Football League) delivers two
    separate XML files, one containing the actual data as well as a metadata file
    containing information about pitch size and start- and endframes of match periods.
    Since no information about framerate is delivered, it is estimated from time
    difference between individual frames. This function provides a high-level access to
    DFL data by parsing "the full match" given both files.

    Parameters
    ----------
    filepath_positions: str or pathlib.Path
        Full path to XML File where the Position data in DFL format is saved.
    filepath_mat_info: str or pathlib.Path
        Full path to XML File where the Match Information data in DFL format is saved.
    links_jID_to_xID: Dict, optional
        A link dictionary of the form `links[team][jID] = xID`. Player's are identified
        in the XML files via jID, and this dictionary is used to map them to a specific
        xID in the respective XY objects. Should be supplied if that order matters. If
        one of links or id_to_jrsy is given as None (default), they are automatically
        extracted from the Match Information XML file.
    links_pID_to_jID: Dict, optional
        A link dictionary of the form `links[team][pID] = jID` where pID is the PersonId
        specified in the DFL Match Information file.

    Returns
    -------
    data_objects: Tuple[XY, XY, XY, XY, XY, XY, Code, Code, Code, Code, Pitch]
        XY-, Code-, and Pitch-objects for both teams and both halves. The order is
        (home_ht1, home_ht2, away_ht1, away_ht2, ball_ht1, ball_ht2,
        possession_ht1, possession_ht2, ballstatus_ht1, ballstatus_ht2, pitch)

    """
    # read metadata
    pitch = read_pitch_from_mat_info_xml(filepath_mat_info)

    # create or check links
    if links_jID_to_xID is None or links_pID_to_jID is None:
        links_jID_to_xID, links_pID_to_jID = create_links_from_mat_info(
            filepath_mat_info
        )
    else:
        pass
        # potential check

    # create periods
    periods, framerate_est = _create_periods_from_dat(filepath_positions)
    segments = list(periods.keys())

    # infer data array shapes
    number_of_home_players = max(links_jID_to_xID["Home"].values()) + 1
    number_of_away_players = max(links_jID_to_xID["Away"].values()) + 1
    number_of_frames = {}
    for segment in segments:
        start = periods[segment][0]
        end = periods[segment][1]
        number_of_frames[segment] = end - start + 1

    # bins
    xydata = {
        "Home": {
            segment: np.full(
                [number_of_frames[segment], number_of_home_players * 2], np.nan
            )
            for segment in segments
        },
        "Away": {
            segment: np.full(
                [number_of_frames[segment], number_of_away_players * 2], np.nan
            )
            for segment in segments
        },
        "Ball": {
            segment: np.full([number_of_frames[segment], 2], np.nan)
            for segment in segments
        },
    }
    codes = {
        code: {segment: [] for segment in segments}
        for code in ["possession", "ballstatus"]
    }

    # loop over frame sets containing player & ball positions for all segments
    for _, frame_set in etree.iterparse(filepath_positions, tag="FrameSet"):

        # ball
        if frame_set.get("TeamId").lower() == "ball":
            # (x, y) position
            segment = frame_set.get("GameSection")
            xydata["Ball"][segment][:, 0] = np.array(
                [float(frame.get("X")) for frame in frame_set.iterfind("Frame")]
            )
            xydata["Ball"][segment][:, 1] = np.array(
                [float(frame.get("Y")) for frame in frame_set.iterfind("Frame")]
            )
            # codes
            codes["ballstatus"][segment] = [
                float(frame.get("BallStatus")) for frame in frame_set.iterfind("Frame")
            ]
            codes["possession"][segment] = [
                float(frame.get("BallPossession"))
                for frame in frame_set.iterfind("Frame")
            ]

        # teams
        else:
            # find identity of frame set
            frames = [frame for frame in frame_set.iterfind("Frame")]
            segment = frame_set.get("GameSection")
            if frame_set.get("PersonId") in links_pID_to_jID["Home"]:
                team = "Home"
                jrsy = links_pID_to_jID[team][frame_set.get("PersonId")]
            elif frame_set.get("PersonId") in links_pID_to_jID["Away"]:
                team = "Away"
                jrsy = links_pID_to_jID[team][frame_set.get("PersonId")]
            else:
                pass
                # possible error or warning

            # insert (x,y) data to correct place in bin
            start = int(frames[0].get("N")) - periods[segment][0]
            end = int(frames[-1].get("N")) - periods[segment][0] + 1
            x_col = (links_jID_to_xID[team][jrsy]) * 2
            y_col = (links_jID_to_xID[team][jrsy]) * 2 + 1
            xydata[team][segment][start:end, x_col] = np.array(
                [float(frame.get("X")) for frame in frames]
            )
            xydata[team][segment][start:end, y_col] = np.array(
                [float(frame.get("Y")) for frame in frames]
            )

        frame_set.clear()

    # create XY objects
    home_ht1 = XY(xy=xydata["Home"]["firstHalf"], framerate=framerate_est)
    home_ht2 = XY(xy=xydata["Home"]["secondHalf"], framerate=framerate_est)
    away_ht1 = XY(xy=xydata["Away"]["firstHalf"], framerate=framerate_est)
    away_ht2 = XY(xy=xydata["Away"]["secondHalf"], framerate=framerate_est)
    ball_ht1 = XY(xy=xydata["Ball"]["firstHalf"], framerate=framerate_est)
    ball_ht2 = XY(xy=xydata["Ball"]["secondHalf"], framerate=framerate_est)

    # create Code objects
    possession_ht1 = Code(
        code=np.array(codes["possession"]["firstHalf"]),
        name="possession",
        definitions={1: "Home", 2: "Away"},
        framerate=framerate_est,
    )
    possession_ht2 = Code(
        code=np.array(codes["possession"]["secondHalf"]),
        name="possession",
        definitions={1: "Home", 2: "Away"},
        framerate=framerate_est,
    )
    ballstatus_ht1 = Code(
        code=np.array(codes["ballstatus"]["firstHalf"]),
        name="ballstatus",
        definitions={0: "Dead", 1: "Alive"},
        framerate=framerate_est,
    )
    ballstatus_ht2 = Code(
        code=np.array(codes["ballstatus"]["secondHalf"]),
        name="ballstatus",
        definitions={0: "Dead", 1: "Alive"},
        framerate=framerate_est,
    )

    data_objects = (
        home_ht1,
        home_ht2,
        away_ht1,
        away_ht2,
        ball_ht1,
        ball_ht2,
        possession_ht1,
        possession_ht2,
        ballstatus_ht1,
        ballstatus_ht2,
        pitch,
    )

    return data_objects



import warnings
from pathlib import Path
from typing import Tuple, Union

import os
import json
import pandas as pd

from floodlight.core.events import Events


def read_open_statsbomb_event_data_json(
    filepath_events: Union[str, Path],
    filepath_match: Union[str, Path],
    filepath_threesixty: Union[str, Path] = None,
) -> Tuple[Events, Events, Events, Events]:
    """Parses files for a single match from the StatsBomb open dataset and extracts the
    event data.

    This function provides a high-level access to an events json file from the openly
    published StatsBomb open data and returns Event objects for both teams for the first
    two periods. A StatsBomb360 json file can be passed to the function to include
    `StatsBomb360 data <https://statsbomb.com/articles/soccer/
    statsbomb-360-freeze-frame-viewer-a-new-release-in-statsbomb-iq/>`_ to the
    ``qualifier`` column. Requires the parsed files from the dataset to maintain their
    original names from the `official data repository
    <https://github.com/statsbomb/open-data>`_

    Parameters
    ----------
    filepath_events: str or pathlib.Path
        Full path to json file where the Event data is saved.
    filepath_match: str or pathlib.Path
        Full path to json file where information about all matches of a season are
        stored.
    filepath_threesixty: str or pathlib.Path, optional
        Full path to json file where the StatsBomb360 data in is saved if available. The
        information about the area of the field where player positions are tracked
        (``visible_area``) and player positions at single events (``freeze frame``) are
        stored as a string in the ``qualifier`` column.

    Returns
    -------
    data_objects: Tuple[Events, Events, Events, Events]
        Events- and Pitch-objects for both teams and both halves. The order is
        (home_ht1, home_ht2, away_ht1, away_ht2).

    Notes
    -----
    StatsBomb's open format of handling provides certain additional event attributes,
    which attach additional information to certain events. As of now, these information
    are parsed as a string in the ``qualifier`` column of the returned DataFrame and can
    be transformed to a dict of form ``{attribute: value}``. This includes the
    information about the tracked position of (some) players and the visible area
    that is included in the StatsBomb360 data.
    """

    # load json files into memory
    with open(filepath_match, "r", encoding="utf8") as f:
        matchinfo_list = json.load(f)
    with open(filepath_events, "r", encoding="utf8") as f:
        file_event_list = json.load(f)
    if filepath_threesixty is not None:
        with open(filepath_threesixty, "r", encoding="utf8") as f:
            file_threesixty_list = json.load(f)
    else:
        file_threesixty_list = None

    # 1. retrieve match info from file
    mID = int(filepath_events.split(os.path.sep)[-1][:-5])  # from filepath
    matchinfo = None
    for info in matchinfo_list:
        if info["match_id"] == mID:
            matchinfo = info
            break

    # 2. parse match info
    teams = ["Home", "Away"]
    tID_link = {
        int(matchinfo["home_team"]["home_team_id"]): "Home",
        int(matchinfo["away_team"]["away_team_id"]): "Away",
    }
    periods = set([event["period"] for event in file_event_list])
    segments = [f"HT{period}" for period in periods]

    # 3. parse events
    # bins
    columns = [
        "eID",
        "gameclock",
        "pID",
        "tID",
        "mID",
        "outcome",
        "timestamp",
        "minute",
        "second",
        "at_x",
        "at_y",
        "to_x",
        "to_y",
        "event_name",
        "player_name",
        "team_name",
        "qualifier",
    ]

    team_event_lists = {
        team: {segment: {col: [] for col in columns} for segment in segments}
        for team in teams
    }

    # loop
    for event in file_event_list:
        # get team and segment information
        period = event["period"]
        segment = "HT" + str(period)
        team = tID_link[event["possession_team"]["id"]]

        # identifier and outcome:
        eID = event["type"]["id"]
        tID = event["team"]["id"]
        pID = event["player"]["id"] if "player" in event else None
        event_name = event["type"]["name"]
        team_name = event["team"]["name"]
        player_name = event["player"]["name"] if "player" in event else None
        outcome = None
        if "type" in event and event["type"]["name"].lower() in event:
            outcome_name = (
                event[event["type"]["name"].lower()]["outcome"]["name"]
                if "outcome" in event[event["type"]["name"].lower()]
                else "None"
            )
            if outcome_name in ["Goal", "Won", "Complete", "Success In Play"]:
                outcome = 1
            elif outcome_name in [
                "Incomplete",
                "Lost In Play",
                "Saved Off Target",
                "Off T",
                "Blocked",
            ]:
                outcome = 0
        team_event_lists[team][segment]["mID"].append(mID)
        team_event_lists[team][segment]["eID"].append(eID)
        team_event_lists[team][segment]["tID"].append(tID)
        team_event_lists[team][segment]["pID"].append(pID)
        team_event_lists[team][segment]["event_name"].append(event_name)
        team_event_lists[team][segment]["team_name"].append(team_name)
        team_event_lists[team][segment]["player_name"].append(player_name)
        team_event_lists[team][segment]["outcome"].append(outcome)

        # relative time
        timestamp = event["timestamp"]
        minute = event["minute"]
        second = event["second"]
        millisecond = int(timestamp.split(".")[1])
        gameclock = 60 * minute + second + millisecond * 0.001
        team_event_lists[team][segment]["timestamp"].append(timestamp)
        team_event_lists[team][segment]["minute"].append(minute)
        team_event_lists[team][segment]["second"].append(second)
        team_event_lists[team][segment]["gameclock"].append(gameclock)

        # location
        at_x = event["location"][0] if "location" in event else None
        at_y = event["location"][1] if "location" in event else None
        if "type" in event and event["type"]["name"].lower() in event:
            to_x = (
                event[event["type"]["name"].lower()]["end_location"][0]
                if "end_location" in event[event["type"]["name"].lower()]
                else None
            )
            to_y = (
                event[event["type"]["name"].lower()]["end_location"][1]
                if "end_location" in event[event["type"]["name"].lower()]
                else None
            )
        else:
            to_x = None
            to_y = None
        team_event_lists[team][segment]["at_x"].append(at_x)
        team_event_lists[team][segment]["at_y"].append(at_y)
        team_event_lists[team][segment]["to_x"].append(to_x)
        team_event_lists[team][segment]["to_y"].append(to_y)

        # qualifier
        qual_dict = {}
        qual_dict["unique_identifier"] = event["id"]
        for qualifier in event:
            if qualifier in [
                "team",
                "player",
                "period",
                "timestamp",
                "minute",
                "second",
                "location",
                "id",
                "type",
            ]:
                continue
            qual_value = event[qualifier]
            qual_dict[qualifier] = qual_value

        if file_threesixty_list is not None:
            threesixty_event = [
                event
                for event in file_threesixty_list
                if event["event_uuid"] == qual_dict["unique_identifier"]
            ]
            if len(threesixty_event) == 1:
                qual_dict["360_freeze_frame"] = threesixty_event[0]["freeze_frame"]
                qual_dict["360_visible_area"] = threesixty_event[0]["visible_area"]
            elif len(threesixty_event) >= 1:
                warnings.warn(
                    f"Found ambiguous StatsBomb event ID "
                    f"{qual_dict['unique_identifier']} matching to more than one "
                    f"StatsBomb360 event."
                )
        team_event_lists[team][segment]["qualifier"].append(str(qual_dict))

    # assembly
    home_ht1 = Events(
        events=pd.DataFrame(data=team_event_lists["Home"]["HT1"]),
    )
    home_ht2 = Events(
        events=pd.DataFrame(data=team_event_lists["Home"]["HT2"]),
    )
    away_ht1 = Events(
        events=pd.DataFrame(data=team_event_lists["Away"]["HT1"]),
    )
    away_ht2 = Events(
        events=pd.DataFrame(data=team_event_lists["Away"]["HT2"]),
    )

    data_objects = (home_ht1, home_ht2, away_ht1, away_ht2)

    return data_objects



from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, List, Tuple, Any, Union
import warnings

import numpy as np
import pandas as pd

from floodlight.utils.types import Numeric
from floodlight.core.definitions import essential_events_columns, protected_columns
from floodlight.core.code import Code


@dataclass
class Events:
    """Event data fragment. Core class of floodlight.

    Event data is stored in `pandas` ``DataFrame``, where each row stores one event
    with its different properties organized in columns. You may put whatever
    information you like in these columns. Yet, the columns `"eID"` and `"gameclock"`
    are mandatory to identify and time-locate events. Some special column names are
    reserved for properties that follow conventions. These may be necessary and their
    existence is checked for running particular analyses.

    Attributes
    ----------
    events: pd.DataFrame
        DataFrame containing rows of events and columns of respective event properties.
    direction: str, optional
        Playing direction of players in data fragment, should be either
        'lr' (left-to-right) or 'rl' (right-to-left).
    essential: list
        List of essential columns available for stored events.
    protected: list
        List of protected columns available for stored events.
    custom: list
        List of custom (i.e. non-essential and non-protected) columns available for
        stored events.
    essential_missing: list or None
        List of missing essential columns or None if no columns are missing.
    essential_invalid: list or None
        List of essential columns that violate the definitions or None if all columns
        match the definitions.
    protected_missing: list or None
        List of missing protected columns or None if no columns are missing.
    protected_invalid: list or None
        List of protected columns that violate the definitions or None if all columns
        match the definitions.
    """

    events: pd.DataFrame
    direction: str = None

    def __post_init__(self):
        # check for missing essential columns
        missing_columns = self.essential_missing
        if missing_columns:
            raise ValueError(
                f"Floodlight Events object is missing the essential "
                f"column(s) {missing_columns}!"
            )

        # warn if value ranges are violated
        incorrect_columns = self.essential_invalid
        if incorrect_columns:
            for col in incorrect_columns:
                warnings.warn(
                    f"Floodlight Events column {col} does not match the defined value"
                    f"range (from floodlight.core.definitions). You can pursue at this "
                    f"point, however, be aware that this may lead to unexpected "
                    f"behavior in the future."
                )

    def __str__(self):
        return f"Floodlight Events object of shape {self.events.shape}"

    def __len__(self):
        return len(self.events)

    def __getitem__(self, key):
        return self.events[key]

    def __setitem__(self, key, value):
        self.events[key] = value

    @property
    def essential(self):
        essential = [
            col for col in self.events.columns if col in essential_events_columns
        ]

        return essential

    @property
    def protected(self):
        protected = [col for col in self.events.columns if col in protected_columns]
        return protected

    @property
    def custom(self):
        custom = [
            col
            for col in self.events.columns
            if col not in essential_events_columns and col not in protected_columns
        ]

        return custom

    @property
    def essential_missing(self):
        missing_columns = [
            col for col in essential_events_columns if col not in self.essential
        ]

        return missing_columns

    @property
    def essential_invalid(self):
        invalid_columns = [
            col
            for col in self.essential
            if not self.column_values_in_range(col, essential_events_columns)
        ]

        return invalid_columns

    @property
    def protected_missing(self):
        missing_columns = [
            col for col in protected_columns if col not in self.protected
        ]

        return missing_columns

    @property
    def protected_invalid(self):
        invalid_columns = [
            col
            for col in self.protected
            if not self.column_values_in_range(col, protected_columns)
        ]

        return invalid_columns

    def column_values_in_range(self, col: str, definitions: Dict[str, Dict]) -> bool:
        """Check if values for a single column of the inner event DataFrame are in
        correct range using using the specifications from floodlight.core.definitions.

        Parameters
        ----------
        col: str
            Column name of the inner event DataFrame to be checked
        definitions: Dict
            Dictionary (from floodlight.core.definitions) containing specifications for
            the columns to be checked.

            The definitions need to contain an entry for the column to be checked and
            this entry needs to contain information about the value range in the form:
            ``definitions[col][value_range] = (min, max)``.

        Returns
        -------
        bool
            True if the checks for value range pass and False otherwise

        Notes
        -----
        Non-integer results of this computation will always be rounded to the next
        smaller integer.
        """
        # skip if value range is not defined
        if definitions[col]["value_range"] is None:
            return True

        # skip values that are None or NaN
        col_nan_free = self.events[col].dropna()

        # retrieve value range from definitions
        min_val, max_val = definitions[col]["value_range"]

        # check value range for remaining values
        if not (min_val <= col_nan_free).all() & (col_nan_free <= max_val).all():
            return False

        # all checks passed
        return True

    def add_frameclock(self, framerate: int):
        """Add the column "frameclock", computed as the rounded multiplication of
        gameclock and framerate, to the inner events DataFrame.

        Parameters
        ----------
        framerate: int
            Temporal resolution of data in frames per second/Hertz.
        """
        frameclock = np.full((len(self.events)), -1, dtype=int)
        frameclock[:] = np.floor(self.events["gameclock"].values * framerate)
        self.events["frameclock"] = frameclock

    def select(
        self, conditions: Tuple[str, Any] or List[Tuple[str, Any]]
    ) -> pd.DataFrame:
        """Returns a DataFrame containing all entries from the inner events DataFrame
         that satisfy all given conditions.

        Parameters
        ----------
        conditions: Tuple or List of Tuples
            A single or a list of conditions used for filtering. Each condition should
            follow the form ``(column, value)``. If ``value`` is given as a variable
            (can also be None), it is used to filter for an exact value. If given as a
            tuple ``value = (min, max)`` that specifies a minimum and maximum value, it
            is filtered for a value range.

            For example, to filter all events that have the ``eID`` of ``"Pass"`` and
            that happened within the first 1000 seconds of the segment, conditions
            should look like:
            ``conditions = [("eID", "Pass"), ("gameclock", (0, 1000))]``

        Returns
        -------
        filtered_events: pd.DataFrame
            A view of the inner events DataFrame with rows fulfilling all criteria
            specified in conditions. The DataFrame can be empty if no row fulfills all
            specified criteria.
        """
        filtered_events = self.events

        # convert single non-list condition to list
        if not isinstance(conditions, list):
            conditions = [conditions]

        # loop through and filter by conditions
        for column, value in conditions:

            # if the value is None filter for all entries with None, NaN or NA
            if value is None:
                filtered_events = filtered_events[filtered_events[column].isna()]

            # check if a single value or a value range is given
            else:

                # value range: filter by minimum and maximum value
                if isinstance(value, (list, tuple)):
                    min_val, max_val = value
                    filtered_events = filtered_events[
                        filtered_events[column] >= min_val
                    ]
                    filtered_events = filtered_events[
                        filtered_events[column] <= max_val
                    ]

                # single value: filter by that value
                else:
                    filtered_events = filtered_events[filtered_events[column] == value]

        return filtered_events

    def translate(self, shift: Tuple[Numeric, Numeric]):
        """Translates data by shift vector.

        Parameters
        ----------
        shift : list or array-like
            Shift vector of form v = (x, y). Any iterable data type with two numeric
            entries is accepted.
        """
        if "at_x" in self.protected and self.events["at_x"].dtype in [
            "int64",
            "float64",
        ]:
            self.events["at_x"] = self.events["at_x"].map(lambda x: x + shift[0])

        if "at_y" in self.protected and self.events["at_y"].dtype in [
            "int64",
            "float64",
        ]:
            self.events["at_y"] = self.events["at_y"].map(lambda x: x + shift[1])

        if "to_x" in self.protected and self.events["to_x"].dtype in [
            "int64",
            "float64",
        ]:
            self.events["to_x"] = self.events["to_x"].map(lambda x: x + shift[0])

        if "to_y" in self.protected and self.events["to_y"].dtype in [
            "int64",
            "float64",
        ]:
            self.events["to_y"] = self.events["to_y"].map(lambda x: x + shift[1])

    def scale(self, factor: float, axis: str = None):
        """Scales data by a given factor and optionally selected axis.

        Parameters
        ----------
        factor : float
            Scaling factor.
        axis : {None, 'x', 'y'}, optional
            Name of scaling axis. If set to 'x' data is scaled on x-axis, if set to 'y'
            data is scaled on y-axis. If None, data is scaled in both directions
            (default).
        """

        if axis not in ["x", "y", None]:
            raise ValueError(f"Expected axis to be one of ('x', 'y', None), got {axis}")

        if axis is None or axis == "x":
            if "at_x" in self.protected and self.events["at_x"].dtype in [
                "int64",
                "float64",
            ]:
                self.events["at_x"] = self.events["at_x"].map(lambda x: x * factor)
            if "to_x" in self.protected and self.events["to_x"].dtype in [
                "int64",
                "float64",
            ]:
                self.events["at_x"] = self.events["at_x"].map(lambda x: x * factor)

        if axis is None or axis == "y":
            if "at_y" in self.protected and self.events["at_y"].dtype in [
                "int64",
                "float64",
            ]:
                self.events["at_y"] = self.events["at_y"].map(lambda x: x * factor)
            if "to_y" in self.protected and self.events["to_y"].dtype in [
                "int64",
                "float64",
            ]:
                self.events["to_y"] = self.events["to_y"].map(lambda x: x * factor)

    def reflect(self, axis: str):
        """Reflects data on given `axis`.

        Parameters
        ----------
        axis : {'x', 'y'}
            Name of reflection axis. If set to "x", data is reflected on x-axis,
            if set to "y", data is reflected on y-axis.
        """
        if axis == "x":
            self.scale(factor=-1, axis="y")
        elif axis == "y":
            self.scale(factor=-1, axis="x")
        else:
            raise ValueError(f"Expected axis to be one of ('x', 'y'), got {axis}")

    def rotate(self, alpha: float):
        """Rotates data on given angle 'alpha' around the origin.

        Parameters
        ----------
        alpha: float
            Rotation angle in degrees. Alpha must be between -360 and 360.
            If positive alpha, data is rotated in counter clockwise direction.
            If negative, data is rotated in clockwise direction around the origin.
        """
        if not (-360 <= alpha <= 360):
            raise ValueError(
                f"Expected alpha to be from -360 to 360, got {alpha} instead"
            )

        phi = np.radians(alpha)
        cos = np.cos(phi)
        sin = np.sin(phi)

        # construct rotation matrix
        r = np.array([[cos, -sin], [sin, cos]]).transpose()

        if "at_x" in self.protected and self.events["at_x"].dtype in [
            "int64",
            "float64",
        ]:
            if "at_y" in self.protected and self.events["at_y"].dtype in [
                "int64",
                "float64",
            ]:
                self.events[["at_x", "at_y"]] = pd.DataFrame(
                    np.round(np.dot(self.events[["at_x", "at_y"]], r), 3)
                )

        if "to_x" in self.protected and self.events["to_x"].dtype in [
            "int64",
            "float64",
        ]:
            if "to_y" in self.protected and self.events["to_y"].dtype in [
                "int64",
                "float64",
            ]:
                self.events[["to_x", "to_y"]] = pd.DataFrame(
                    np.round(np.dot(self.events[["to_x", "to_y"]], r), 3)
                )

    def slice(
        self,
        start: float = None,
        end: float = None,
        slice_by="gameclock",
        inplace: bool = False,
    ):
        """Return copy of object with events sliced in a time interval.

        Intended columns for using this function are ``gameclock`` (total seconds) or
        ``frameclock``. However, also allows slicing by any other column that manifests
        a  temporal relation between events (e.g. ``minute``). Excludes all entries
        without a valid entry in the specified column (e.g. None).

        Parameters
        ----------
        start : float, optional
            Start frame or second of slice. Defaults to beginning of segment.
        end : float, optional
            End frame or second of slice (endframe is excluded). Defaults to last event
            of segment (including).
        slice_by: {'gameclock', 'frameclock'}, optional
            Column used to slice the events. Defaults to ``gameclock``.
        inplace: bool, optional
            If set to ``False`` (default), a new object is returned, otherwise the
            operation is performed in place on the called object.

        Returns
        -------
        events_sliced: Union[Event, None]
        """
        if slice_by not in self.events:
            ValueError(f"Events object does not contain column {slice_by}!")
        if start is None:
            start = 0
        if end is None:
            end = np.nanmax(self.events[slice_by].values) + 1

        sliced_data = self.events[self.events[slice_by] >= start].copy()
        sliced_data = sliced_data[sliced_data[slice_by] < end]
        sliced_data.reset_index(drop=True, inplace=True)

        events_sliced = None
        if inplace:
            self.events = sliced_data
        else:
            events_sliced = Events(
                events=sliced_data,
                direction=deepcopy(self.direction),
            )

        return events_sliced

    def get_event_stream(
        self,
        fade: Union[int, None] = 0,
        **kwargs,
    ) -> Code:
        """Generates a Code object containing the eIDs of all events at the
        respective frame and optionally subsequent frames as defined by the fade
        argument.

        This function translates the object's DataFrame of temporally irregular events
        to a continuous frame-wise representation. This can be especially helpful to
        connect event data with spatiotemporal data, e.g., for filtering the latter
        based on the former. Events overwrite preceding event's fade, and unfilled
        values are set to np.nan.

        Notes
        ------
        Requires the DataFrame to contain the protected ``frameclock`` column.

        Parameters
        ----------
        fade: int, optional
            Number of additional frames for which the Code object should stay at a
            value after the event occurred. The value is overwritten if another event
            occurs within the fade duration. If chosen to zero, the value is maintained
            only for a single frame. If chosen to None, the value is maintained until
            either the next event or until the end of the sequence. Defaults to 0.
        kwargs:
            Keyword arguments of the Code object ("name", "definitions", "framerate")
            that are passed down to instantiate the returned event_stream.

        Returns
        -------
        event_stream: Code
            Generated continuous event stream describing the designated game state.
        """
        if "frameclock" in self.protected_missing:
            raise ValueError(
                "Cannot create event stream from Events object missing "
                "the protected column 'frameclock'. Consider calling "
                "add_frameclock to the Events object first!"
            )
        if fade is not None and fade < 0:
            raise ValueError(
                f"Expected fade to be a positive integer or None, got {fade} instead."
            )

        sorted_events = self.events.sort_values("frameclock")
        start = int(np.round(np.nanmin(sorted_events["frameclock"].values)))
        end = int(np.round(np.nanmax(sorted_events["frameclock"].values))) + 1

        code = np.full((end - start,), np.nan, dtype=object)
        for _, event in sorted_events.iterrows():
            if pd.isna(event["frameclock"]):
                continue
            frame = int(np.round(event["frameclock"]))
            if fade is None:
                code[frame - start :] = event["eID"]
            else:
                code[frame - start : frame - start + fade + 1] = event["eID"]

        event_stream = Code(
            code=code,
            name=kwargs.get("name"),
            definitions=kwargs.get("definitions"),
            framerate=kwargs.get("framerate"),
        )

        return event_stream



from pathlib import Path
from typing import Dict, Tuple, Union

import numpy as np
from lxml import etree

from floodlight.core.code import Code
from floodlight.core.pitch import Pitch
from floodlight.core.xy import XY


def _read_metadata(filepath_metadata: Union[str, Path]) -> Tuple[Dict, Dict, Pitch]:
    """Reads TRACAB's metadata file and extracts information about match metainfo,
    periods and the pitch.

    Parameters
    ----------
    filepath_metadata: str or pathlib.Path
        Full path to metadata.xml file.

    Returns
    -------
    metainfo: Dict
        Dictionary with metainformation such as framerate.
    periods: Dict
        Dictionary with start and endframes:
        `periods[segment] = (startframe, endframe)`.
    pitch: Pitch
        Pitch object with actual pitch length and width.
    """
    #  set up XML tree
    tree = etree.parse(str(filepath_metadata))
    root = tree.getroot()

    # parse XML file, extract matchinfo and period start/endframes
    metadata = {}
    periods = {}
    attributes = root.find("match").attrib

    framerate = attributes.get("iFrameRateFps")
    metadata["framerate"] = int(framerate) if framerate else None

    length = attributes.get("fPitchXSizeMeters")
    metadata["length"] = float(length) if length else None

    width = attributes.get("fPitchYSizeMeters")
    metadata["width"] = float(width) if width else None

    for elem in root.findall("match/period"):
        if elem.attrib["iEndFrame"] != "0":
            segment = "HT" + elem.attrib["iId"]
            start = int(elem.attrib["iStartFrame"])
            end = int(elem.attrib["iEndFrame"])
            periods[segment] = (start, end)

    pitch = Pitch.from_template(
        "tracab",
        length=float(metadata["length"]),
        width=float(metadata["width"]),
        sport="football",
    )

    return metadata, periods, pitch


def _read_dat_single_line(
    package: str,
) -> Tuple[
    int, Dict[str, Dict[str, Tuple[float, float, float]]], Dict[str, Union[str, tuple]]
]:
    """Extracts all relevant information from a single line of TRACAB's .dat file
    (i.e. one frame of data).

    Parameters
    ----------
    package: str
        One full line from TRACAB's .dat-file, equals one "package" according to the
        file-format documentation.

    Returns
    -------
    frame_number: int
        The number of current frame.
    positions: Dict[str, Dict[str, Tuple[float, float, float]]]
        Nested dictionary that stores player position information for each team and
        player. Has the form `positions[team][jID] = (x, y, speed)`.
    ball: Dict[str]
        Dictionary with ball information. Has keys 'position', 'possession' and
        'ballstatus'.
    """
    # bins
    positions = {"Home": {}, "Away": {}, "Other": {}}
    ball = {}

    # split package to chunks
    chunk1, chunk2, chunk3, _ = package.split(sep=":")

    # first chunk (frame number)
    frame_number = int(chunk1)

    # second chunk (player positions)
    targets = chunk2[:-1].split(sep=";")
    for t in targets:
        player_data = t.split(sep=",")
        # type conversions
        team, system_id, jID = map(lambda x: int(x), player_data[:3])
        x, y, speed = map(lambda x: float(x), player_data[3:])
        if team == 1:
            team = "Home"
        elif team == 0:
            team = "Away"
        else:
            team = "Other"
        # assign
        positions[team][jID] = (x, y, speed)

    # third chunk (ball data)
    ball_data = chunk3.split(sep=",")[:6]
    ball["position"] = tuple(map(lambda x: float(x), ball_data[:2]))
    ball["possession"] = ball_data[4]
    ball["ballstatus"] = ball_data[5][0]

    return frame_number, positions, ball


def _frame_in_period(
    frame_number: int, periods: Dict[str, Tuple[int, int]]
) -> Union[str, None]:
    """Checks if a given frame is within the range of start- and endframe for all
    periods and returns the name of the period the frame belongs to, or None if it
    can't find any.

    Parameters
    ----------
    frame_number: int
        Frame number to be checked.
    periods: Dict[str, Tuple[int, int]]
        Dictionary with period start- and endframes of the form
        `periods[segment] = (startframe, endframe)` as it is returned by
        :meth:`floodlight.io.tracab._read_metadata`.

    Returns
    -------
    segment: str or None
        Name of the segment the frame belongs to, or None if it does not belong to any
        of the supplied segments.
    """
    # determine current segment by iterating through all segments (i)
    segment = None
    for i in periods.keys():
        if frame_number in range(periods[i][0], periods[i][1] + 1):
            segment = i

    return segment


def _read_dat_jersey_numbers(filepath_dat: Union[str, Path]):
    """Reads entire TRACAB .dat file and extracts unique set of jIDs (jerseynumbers)
    for both teams.

    Parameters
    ----------
    filepath_dat: str or pathlib.Path
        Full path to .dat file.

    Returns
    -------
    home_jIDs: set
    away_jIDs: set
    """
    # bins
    home_jIDs = set()
    away_jIDs = set()
    # loop
    with open(str(filepath_dat), "r") as f:
        while True:
            package = f.readline()
            # terminate if at end of file
            if len(package) == 0:
                break
            # read line
            _, positions, _ = _read_dat_single_line(package)
            # Extract jersey numbers
            home_jIDs |= positions["Home"].keys()
            away_jIDs |= positions["Away"].keys()

    return home_jIDs, away_jIDs


def create_links_from_dat(filepath_dat: Union[str, Path]) -> Dict[str, Dict[int, int]]:
    """Parses the entire TRACAB .dat file for unique jIDs (jerseynumbers) and creates a
    dictionary linking jIDs to xIDs in ascending order.

    Parameters
    ----------
    filepath_dat: str or pathlib.Path
        Full path to .dat file.

    Returns
    -------
    links: Dict[str, Dict[int, int]]
        Link-dictionary of the form `links[team][jID] = xID`.
    """
    homejrsy, awayjrsy = _read_dat_jersey_numbers(filepath_dat)

    homejrsy = list(homejrsy)
    awayjrsy = list(awayjrsy)

    homejrsy.sort()
    awayjrsy.sort()

    links = {
        "Home": {jID: xID for xID, jID in enumerate(homejrsy)},
        "Away": {jID: xID for xID, jID in enumerate(awayjrsy)},
    }

    return links


def read_tracab_files(
    filepath_dat: Union[str, Path],
    filepath_metadata: Union[str, Path],
    links: Dict[str, Dict[int, int]] = None,
) -> Tuple[XY, XY, XY, XY, XY, XY, Code, Code, Code, Code, Pitch]:
    """Parse TRACAB files and extract position data, possession and ballstatus codes as
    well as pitch information.

    ChyronHego's TRACAB system delivers two separate files, a .dat file containing the
    actual data as well as a metadata.xml containing information about pitch size,
    framerate and start- and endframes of match periods. This function provides a
    high-level access to TRACAB data by parsing "the full match" given both files.

    Parameters
    ----------
    filepath_dat: str or pathlib.Path
        Full path to dat-file.
    filepath_metadata: str or pathlib.Path
        Full path to metadata.xml file.
    links: Dict[str, Dict[int, int]], optional
        A link dictionary of the form `links[team][jID] = xID`. Player's are identified
        in TRACAB files via jID, and this dictionary is used to map them to a specific
        xID in the respective XY objects. Should be supplied if that order matters. If
        None is given (default), the links are automatically extracted from the .dat
        file at the cost of a second pass through the entire file.

    Returns
    -------
    data_objects: Tuple[XY, XY, XY, XY, XY, XY, Code, Code, Code, Code, Pitch]
        XY-, Code-, and Pitch-objects for both teams and both halves. The order is
        (home_ht1, home_ht2, away_ht1, away_ht2, ball_ht1, ball_ht2,
        possession_ht1, possession_ht2, ballstatus_ht1, ballstatus_ht2, pitch)
    """
    # read metadata
    metadata, periods, pitch = _read_metadata(filepath_metadata)
    segments = list(periods.keys())

    # create or check links
    if links is None:
        links = create_links_from_dat(filepath_dat)
    else:
        pass
        # potential check vs jerseys in dat file

    # infer data array shapes
    number_of_home_players = max(links["Home"].values()) + 1
    number_of_away_players = max(links["Away"].values()) + 1
    number_of_frames = {}
    for segment in segments:
        start = periods[segment][0]
        end = periods[segment][1]
        number_of_frames[segment] = end - start + 1

    # bins
    xydata = {}
    xydata["Home"] = {
        segment: np.full(
            [number_of_frames[segment], number_of_home_players * 2], np.nan
        )
        for segment in segments
    }
    xydata["Away"] = {
        segment: np.full(
            [number_of_frames[segment], number_of_away_players * 2], np.nan
        )
        for segment in segments
    }
    xydata["Ball"] = {
        segment: np.full([number_of_frames[segment], 2], np.nan) for segment in segments
    }
    codes = {
        code: {segment: [] for segment in segments}
        for code in ["possession", "ballstatus"]
    }

    # loop
    with open(filepath_dat, "r") as f:
        while True:
            package = f.readline()
            # terminate if at end of file
            if len(package) == 0:
                break
            # read line to get absolute frame (in file), player positions and ball info
            frame_abs, positions, ball = _read_dat_single_line(package)

            # check if frame is in any segment
            segment = _frame_in_period(frame_abs, periods)
            if segment is None:
                # skip line if not
                continue
            else:
                # otherwise calculate relative frame (in respective segment)
                frame_rel = frame_abs - periods[segment][0]

            # insert (x,y)-data into correct np.array, at correct place (t, xID)
            for team in ["Home", "Away"]:
                for jID in positions[team].keys():
                    # map jersey number to array index and infer respective columns
                    x_col = (links[team][jID]) * 2
                    y_col = (links[team][jID]) * 2 + 1
                    xydata[team][segment][frame_rel, x_col] = positions[team][jID][0]
                    xydata[team][segment][frame_rel, y_col] = positions[team][jID][1]

            # get ball data
            xydata["Ball"][segment][
                frame_rel,
            ] = ball["position"]
            codes["possession"][segment].append(ball.get("possession", np.nan))
            codes["ballstatus"][segment].append(ball.get("ballstatus", np.nan))

    # create XY objects
    home_ht1 = XY(xy=xydata["Home"]["HT1"], framerate=metadata["framerate"])
    home_ht2 = XY(xy=xydata["Home"]["HT2"], framerate=metadata["framerate"])
    away_ht1 = XY(xy=xydata["Away"]["HT1"], framerate=metadata["framerate"])
    away_ht2 = XY(xy=xydata["Away"]["HT2"], framerate=metadata["framerate"])
    ball_ht1 = XY(xy=xydata["Ball"]["HT1"], framerate=metadata["framerate"])
    ball_ht2 = XY(xy=xydata["Ball"]["HT2"], framerate=metadata["framerate"])

    # create Code objects
    possession_ht1 = Code(
        code=np.array(codes["possession"]["HT1"]),
        name="possession",
        definitions={"H": "Home", "A": "Away"},
        framerate=metadata["framerate"],
    )
    possession_ht2 = Code(
        code=np.array(codes["possession"]["HT2"]),
        name="possession",
        definitions={"H": "Home", "A": "Away"},
        framerate=metadata["framerate"],
    )
    ballstatus_ht1 = Code(
        code=np.array(codes["ballstatus"]["HT1"]),
        name="ballstatus",
        definitions={"D": "Dead", "A": "Alive"},
        framerate=metadata["framerate"],
    )
    ballstatus_ht2 = Code(
        code=np.array(codes["ballstatus"]["HT2"]),
        name="ballstatus",
        definitions={"D": "Dead", "A": "Alive"},
        framerate=metadata["framerate"],
    )

    data_objects = (
        home_ht1,
        home_ht2,
        away_ht1,
        away_ht2,
        ball_ht1,
        ball_ht2,
        possession_ht1,
        possession_ht2,
        ballstatus_ht1,
        ballstatus_ht2,
        pitch,
    )

    return data_objects



import warnings
from pathlib import Path
from typing import List, Dict, Tuple, Union

import numpy as np

from floodlight.core.xy import XY


def get_column_names_from_csv(filepath_data: Union[str, Path]) -> List[str]:
    """Reads first line of a Kinexon.csv-file and extracts the column names.

    Parameters
    ----------
    filepath_data: str or pathlib.Path
        Full path to Kinexon.csv-file.

    Returns
    -------
    columns: List[str]
        List with every column name of the .csv-file.
    """

    with open(str(filepath_data)) as f:
        columns = f.readline().split(",")

    return columns


def _get_column_links(filepath_data: Union[str, Path]) -> Union[None, Dict[str, int]]:
    """Creates a dictionary with the relevant recorded columns and their
    corresponding column index in the Kinexon.csv-file.

    Parameters
    ----------
    filepath_data: str or pathlib.Path
        Full path to Kinexon.csv-file.

    Returns
    -------
    column_links: Dict[str, int]
        Dictionary with column index for relevant recorded columns.
        'column_links[column] = index'
        The following columns are currently considered relevant:
              floodlight id: 'column name in Kinexon.csv-file
            - time: 'ts in ms'
            - sensor_id: 'sensor id'
            - mapped_id: 'mapped id'
            - name: 'full name'
            - group_id: 'group id'
            - x_coord: 'x in m'
            - y_coord: 'y in m'
    """

    recorded_columns = get_column_names_from_csv(str(filepath_data))

    # relevant columns
    mapping = {
        "ts in ms": "time",
        "sensor id": "sensor_id",
        "mapped id": "mapped_id",
        "full name": "name",
        "number": "number",
        "group id": "group_id",
        "group name": "group_name",
        "x in m": "x_coord",
        "y in m": "y_coord",
    }

    necessary_columns = ["time", "x_coord", "y_coord"]

    column_links = {}
    # loop
    for key in mapping:
        # create links
        if key in recorded_columns:
            column_links.update({mapping[key]: recorded_columns.index(key)})

    # check if necessary columns are available
    if not all(columns in column_links for columns in necessary_columns):
        warnings.warn(
            "Data file lacks critical information! "
            "No timestamp or coordinates found."
        )
        return None

    return column_links


def _get_group_id(
    recorded_group_identifier: List[str],
    column_links: Dict[str, int],
    single_line: List[str],
) -> Union[str, None]:
    """Returns the group_name or group_id if it was recorded or "0" if not.
    Favors the group_name over the group_id.

    Parameters
    ----------
    recorded_group_identifier: List[str]
        List of all recorded group identifiers. Group identifiers are "group_id" or
        "group_name".
    column_links: Dict[str, int]
        Dictionary with column index for relevant recorded columns.
        'column_links[column] = index'
        The following columns are currently considered relevant:
              floodlight id: 'column name in Kinexon.csv-file
            - time: 'ts in ms'
            - sensor_id: 'sensor id'
            - mapped_id: 'mapped id'
            - name: 'full name'
            - group_id: 'group id'
            - x_coord: 'x in m'
            - y_coord: 'y in m'
    single_line: List[str]
        Single line of a Kinexon.csv-file that has been split at the respective
        delimiter, eg. ",".

    Returns
    -------
    group_id: str
        The respective group id in that line or "0" if there is no group id.
    """

    # check for group identifier
    has_groups = len(recorded_group_identifier) > 0

    if has_groups:
        # extract group identifier
        if "group_name" in recorded_group_identifier:
            group_identifier = "group_name"
        elif "group_id" in recorded_group_identifier:
            group_identifier = "group_id"
        else:
            warnings.warn("Data has groups but no group identifier!")
            return None

        group_id = single_line[column_links[group_identifier]]

    # no groups
    else:
        group_id = "0"

    return group_id


def get_meta_data(
    filepath_data: Union[str, Path]
) -> Tuple[Dict[str, Dict[str, List[str]]], int, int, int]:
    """Reads Kinexon's position data file and extracts meta-data about groups, sensors,
    length and framerate.

    Parameters
    ----------
    filepath_data: str or pathlib.Path
        Full path to Kinexon.csv-file.

    Returns
    -------
    pID_dict: Dict[str, Dict[str, List[str]]],
        Nested dictionary that stores information about the pIDs from every player-
        identifying column in every group.
        'pID_dict[group_identifier][identifying_column] = [pID1, pID2, ..., pIDn]'
        When recording and exporting Kinexon data, the pID can be stored
        in different columns. Player-identifying columns are "sensor_id", "mapped_id",
        and "full_name". If the respective column is in the recorded data, its pIDs are
        listed in pID_dict.
        As with pID, group ids can be stored in different columns. Group-identifying
        columns are "group_name" and "group_id". If both are available, group_name will
        be favored over group_id as the group_identifier.
    number_of_frames: int
        Number of frames from the first to the last recorded frame.
    framerate: int
        Estimated framerate in frames per second. Estimated from the smallest difference
        between two consecutive frames.
    t_null: int
        Timestamp of the first recorded frame
    """

    column_links = _get_column_links(str(filepath_data))
    sensor_identifier = {"name", "number", "sensor_id", "mapped_id"}
    column_links_set = set(column_links)
    recorded_sensor_identifier = list(column_links_set & sensor_identifier)
    sensor_links = {
        key: index
        for (key, index) in column_links.items()
        if key in recorded_sensor_identifier
    }
    group_identifier_set = {"group_id", "group_name"}
    recorded_group_identifier = list(column_links_set & group_identifier_set)

    # dict for pIDs
    pID_dict = {}
    # list for timestamps
    t = []
    # check for group identifier
    has_groups = len(recorded_group_identifier) > 0
    if not has_groups:
        warnings.warn("Since no group exist in data, dummy group '0' is created!")

    # loop
    with open(str(filepath_data), "r") as f:
        # skip the header of the file
        _ = f.readline()
        while True:
            line_string = f.readline()
            # terminate if at end of file
            if len(line_string) == 0:
                break
            # split str
            line = line_string.split(",")
            # extract frames timestamp
            t.append(int(line[column_links["time"]]))
            # extract group_id
            group_id = _get_group_id(recorded_group_identifier, column_links, line)
            # create group dict in pID_dict
            if group_id not in pID_dict:
                pID_dict.update({group_id: {}})
            # create links
            for identifier in sensor_links:
                # extract identifier
                if identifier not in pID_dict[group_id]:
                    pID_dict[group_id].update({identifier: []})
                # extract ids
                if line[column_links[identifier]] not in pID_dict[group_id][identifier]:
                    pID_dict[group_id][identifier].append(
                        line[column_links[identifier]]
                    )

    # sort dict
    pID_dict = dict(sorted(pID_dict.items()))

    # estimate framerate
    timestamps = list(set(t))
    timestamps.sort()
    timestamps = np.array(timestamps)
    minimum_time_step = np.min(np.diff(timestamps))
    # timestamps are in milliseconds. Magic number 1000 is needed for conversion to
    # seconds.
    framerate = 1000 / minimum_time_step

    # non-integer framerate
    if not framerate.is_integer():
        warnings.warn(
            f"Non-integer frame rate: Minimum time step of "
            f"{minimum_time_step} detected. Framerate is round to "
            f"{int(framerate)}."
        )

    framerate = int(framerate)

    # 1000 again needed to account for millisecond to second conversion.
    number_of_frames = int((timestamps[-1] - timestamps[0]) / (1000 / framerate))
    t_null = timestamps[0]

    return pID_dict, number_of_frames, framerate, t_null


def _get_available_sensor_identifier(pID_dict: Dict[str, Dict[str, List[str]]]) -> str:
    """Returns an available sensor identifier that has been recorded. Will favor
    "name" over "mapped_id" over "sensor_id" over "number".

    Parameters
    ----------
    pID_dict: Dict[str, Dict[str, List[str]]],
        Nested dictionary that stores information about the pIDs from every player-
        identifying column in every group.
        'pID_dict[group][identifying_column] = [pID1, pID2, ..., pIDn]'
        When recording and exporting Kinexon data, the pID can be stored
        in different columns. Player-identifying columns are "sensor_id", "mapped_id",
        and "full_name". If the respective column is in the recorded data, its pIDs are
        listed in pID_dict.

    Returns
    -------
    identifier: str
        One sensor identifier that has been recorded.
    """

    player_identifiers = ["name", "mapped_id", "sensor_id", "number"]
    available_identifier = [
        idt for idt in player_identifiers if idt in list(pID_dict.values())[0]
    ]
    identifier = available_identifier[0]

    return identifier


def create_links_from_meta_data(
    pID_dict: Dict[str, Dict[str, List[str]]], identifier: str = None
) -> Dict[str, Dict[str, int]]:
    """Creates a dictionary from the pID_dict linking the identifier to the xID.

    Parameters
    ----------
    pID_dict: Dict[str, Dict[str, List[str]]]
        Nested dictionary that stores information about the pIDs from every player-
        identifying column in every group. The format is
        ``pID_dict[group][identifying_column] = [pID1, pID2, ..., pIDn]``.
        When recording and exporting Kinexon data, the pID can be stored in different
        columns. Player-identifying columns are ``"sensor_id"``, ``"mapped_id"``, and
        ``"full_name"``. If the respective column is in the recorded data, its pIDs are
        listed in ``pID_dict``.
    identifier: str, optional
        Column-name of personal identifier in Kinexon.csv-file, defaults to None.
        Can be one of: ``"sensor_id"``, ``"mapped_id"``, ``"name"``.

        When recording and exporting Kinexon data, the pID can be stored
        in different columns. Player-identifying columns are ``"sensor_id"``,
        ``"mapped_id"``, and ``"full_name"``. If specified to one of the above, keys in
        links will be the pIDs in that column. If not specified, it will use one of the
        columns, favoring ``"name"`` over ``"mapped_id"`` over ``"sensor_id"``.

    Returns
    -------
    links: Dict[str, Dict[str, int]]
        Link-dictionary of the form ``links[group][identifier-ID] = xID``.
    """

    if identifier is None:
        # available sensor identifier
        identifier = _get_available_sensor_identifier(pID_dict)

    links = {}
    for group_id in pID_dict:
        links.update(
            {
                group_id: {
                    ID: xID for xID, ID in enumerate(pID_dict[group_id][identifier])
                }
            }
        )

    return links


def read_kinexon_file(filepath_data: Union[str, Path]) -> List[XY]:
    """Parse Kinexon files and extract position data.

    Kinexon's local positioning system delivers one .csv file containing the position
    data. This function provides a high-level access to Kinexon data by parsing "the
    full file" given the path to the file.

    Parameters
    ----------
    filepath_data: str or pathlib.Path
        Full path to Kinexon.csv-file.

    Returns
    -------
    positions: List[XY]
        List of XY-objects for the whole game, one per group. The order of groups is
        ascending according to their group_id. If no groups are specified in the file,
        all data gets assigned to a dummy group "0". The order inside the groups is
        ascending according to their appearance in the data.
    """

    # get metadata
    pID_dict, number_of_frames, framerate, t_null = get_meta_data(filepath_data)

    # get links
    links = create_links_from_meta_data(pID_dict)
    # get column-links
    column_links = _get_column_links(filepath_data)
    column_links_set = set(column_links)
    group_identifier_set = {"group_id", "group_name"}
    recorded_group_identifier = list(column_links_set & group_identifier_set)

    # available sensor identifier
    identifier = _get_available_sensor_identifier(pID_dict)

    # create np.array for data
    number_of_sensors = {}
    xydata = {}
    for group in links:
        number_of_sensors.update({group: len(links[group])})
        xydata.update(
            {
                group: np.full(
                    [number_of_frames + 1, number_of_sensors[group] * 2], np.nan
                )
            }
        )

    # loop
    with open(str(filepath_data), "r") as f:
        # skip the header of the file
        _ = f.readline()
        while True:
            line_string = f.readline()
            # terminate if at end of file
            if len(line_string) == 0:
                break
            # split str
            line = line_string.split(",")
            timestamp = int(line[column_links["time"]])
            # set group
            group_id = _get_group_id(recorded_group_identifier, column_links, line)
            # set column
            x_col = links[group_id][line[column_links[identifier]]] * 2
            y_col = x_col + 1
            # set row
            row = int((timestamp - t_null) / (1000 / framerate))
            # set (x, y)-data
            x_coordinate = line[column_links["x_coord"]]
            y_coordinate = line[column_links["y_coord"]]

            # insert (x, y)-data into column and row of respective array
            if x_coordinate != "":
                xydata[group_id][row, x_col] = x_coordinate

            if y_coordinate != "":
                xydata[group_id][row, y_col] = y_coordinate

    data_objects = []
    for group_id in xydata:
        data_objects.append(XY(xy=xydata[group_id], framerate=framerate))

    return data_objects



from functools import wraps

import matplotlib.pyplot as plt


def check_axes_given(func):
    """Decorator function that checks if a matplotlib.axes is given as an argument.
    Creates one if not.

    Parameters
    ----------
    func:
        Function object that needs a matplotlib.axes as an argument. If ax == None
        an axes is created an passed to the given function object as a keyworded
        argument.

    Returns
    -------
    func:
        Function with matplotlib.axes as additional argument if not specified.
        Otherwise the function is returned as it is.
    """

    @wraps(func)
    def add_ax(*args, **kwargs):  # actual wrapper function that gets args and kwargs
        # from the funtion that was passed
        # If matplotlib.axes is not given (ax == None) an axes is created.
        if not kwargs.get("ax"):
            kwargs.pop("ax")  # Remove ax from kwargs
            ax = plt.subplots()[1]  # Create matplotlib.axes
            return func(*args, ax=ax, **kwargs)  # return function with axes

        # If matplotlib.axes is given nothing changes and the function is returned with
        # the given *args and **kwargs
        return func(*args, **kwargs)

    return add_ax



import matplotlib

from floodlight.vis.utils import check_axes_given


@check_axes_given
def plot_positions(
    xy, frame: int, ball: bool, ax: matplotlib.axes, **kwargs
) -> matplotlib.axes:
    """Scatter plots positions for a given frame of an XY object on a
    matplotlib.axes.

    Parameters
    ----------
    xy: floodlight.core.xy.XY
        XY object containing spatiotemporal data to be plotted.
    frame: int
        Number of frame to be plotted.
    ball: bool
         Boolean indicating whether this object is storing ball data. If set to False
         marker="o", else marker=".".
    ax: matplotlib.axes
        Axes from matplotlib library on which the positions are plotted.
    kwargs:
        Optional keyworded arguments e.g. {'color', 'zorder', 'marker'}
        which can be used for the plot functions from matplotlib. The kwargs are
        only passed to the plot functions of matplotlib.

    Returns
    -------
    axes: matplotib.axes
        Axes from matplotlib library on which the positions are plotted.
    """

    # kwargs which are used to configure the plot with default values.
    # All other kwargs are just getting passed to the scatter() method.
    marker = kwargs.pop("marker", "o" if not ball else ".")
    color = kwargs.pop("color", "black" if not ball else "grey")
    zorder = kwargs.pop("zorder", 1)

    # plotting the positions
    # if ball is false
    if not ball:
        ax.scatter(
            x=xy.x[frame],
            y=xy.y[frame],
            marker=marker,
            color=color,
            zorder=zorder,
            **kwargs,
        )
    # if ball is true
    elif ball:
        ax.scatter(
            x=xy.x[frame],
            y=xy.y[frame],
            marker=marker,
            color=color,
            zorder=zorder,
            **kwargs,
        )

    return ax


@check_axes_given
def plot_trajectories(
    xy,
    start_frame: int,
    end_frame: int,
    ball: bool,
    ax: matplotlib.axes,
    **kwargs,
) -> matplotlib.axes:
    """Draws the trajectories of an XY object from a given time interval on a
    matplotlib.axes.

    Parameters
    ----------
    xy: floodlight.core.xy.XY
         XY object containing spatiotemporal data to be plotted.
    start_frame: int
        Starting frame of time interval to be plotted.
    end_frame: int
        Closing frame of time interval to be plotted.
    ball: bool
        Boolean indicating whether this object is storing ball data. If set to False
        marker="o", else marker=".".
    ax: matplotlib.axes
        Axes from matplotlib library on which the trajectories are drawn.
    kwargs:
        Optional keyworded arguments e.g. {'linewidth', 'zorder', 'linestyle', 'alpha'}
        which can be used for the plot functions from matplotlib. The kwargs are only
        passed to all the plot functions of matplotlib.

    Returns
    -------
    axes: matplotib.axes
        Axes from matplotlib library on which the trajectories are drawn.
    """

    # kwargs which are used to configure the plot with default values.
    # All other kwargs are just getting passed to the plot() method.
    color = kwargs.pop("color", "black" if not ball else "grey")
    zorder = kwargs.pop("zorder", 1)
    # ball trajectories are thinner per default
    linewidth = kwargs.pop("linewidth", 1 if not ball else 0.5)

    # iterating over every object (for instance players) in the XY.xy array and plot the
    # trajectories for the given range of frames
    for i in range(0, xy.N, 2):

        x = xy.xy[start_frame:end_frame, i]
        y = xy.xy[start_frame:end_frame, i + 1]
        ax.plot(x, y, color=color, zorder=zorder, linewidth=linewidth, **kwargs)

    return ax



import warnings
from functools import wraps

from floodlight import Pitch


class BaseModel:
    """Base class for all models.

    Attributes
    ----------
    pitch: Pitch, optional
        Some models require pitch information, so the corresponding Pitch object is
        handled during initialization.
    """

    def __init__(self, pitch: Pitch = None):

        if pitch is not None:
            self.check_pitch(pitch)
            self._pitch = pitch

    def __str__(self):
        return f"Floodlight {self.__class__.__name__}"

    @property
    def is_fitted(self) -> bool:
        """Returns ``True`` if all model parameters (those with a trailing underscore)
        are fitted (i.e. not None), and ``False`` otherwise."""
        fitted = all(
            [
                vars(self)[v] is not None
                for v in vars(self)
                if (v.endswith("_") and not v.startswith("__"))
            ]
        )

        return fitted

    @staticmethod
    def check_pitch(pitch: Pitch):
        """
        Performs a series of checks on a Pitch object and raises warnings if the pitch
        configuration may affect computation results.

        Parameters
        ----------
        pitch: Pitch
            Pitch object the checks are performed on.
        """
        # check if metrical system
        if not pitch.is_metrical:
            warnings.warn(
                "Model initialized with non-metrical pitch. Results may be distorted, "
                "use at your own risk.",
                category=RuntimeWarning,
            )


def requires_fit(func):
    """Decorator function for Model-based class-methods that require a previous call to
    that model's fit()-method. Raises a ValueError if fit() has not been called yet."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        model = args[0]
        if not model.is_fitted:
            raise ValueError(
                f"Not all model parameters have been calculated yet. Try "
                f"running {model.__class__.__name__}.fit() before calling "
                f"this method"
            )
        return func(*args, **kwargs)

    return wrapper
